import ast
from io import StringIO

from labelbox import parser


def test_loads(ndjson_content):
    expected_line, expected_objects = ndjson_content
    parsed_line = parser.loads(expected_line)

    assert parsed_line == expected_objects
    assert parser.dumps(parsed_line) == expected_line


def test_loads_bytes(ndjson_content):
    expected_line, expected_objects = ndjson_content

    bytes_line = expected_line.encode("utf-8")
    parsed_line = parser.loads(bytes_line)

    assert parsed_line == expected_objects
    assert parser.dumps(parsed_line) == expected_line


def test_reader_stringio(ndjson_content):
    line, ndjson_objects = ndjson_content

    text_io = StringIO(line)
    parsed_arr = []
    reader = parser.reader(text_io)
    for _, r in enumerate(reader):
        parsed_arr.append(r)
    assert parsed_arr == ndjson_objects


def test_non_ascii_new_line(ndjson_content_with_nonascii_and_line_breaks):
    line, expected_objects = ndjson_content_with_nonascii_and_line_breaks
    parsed = parser.loads(line)

    assert parsed == expected_objects

    # NOTE: json parser converts unicode chars to unicode literals by default and this is a good practice
    #   but it is not what we want here since we want to compare the strings with actual unicode chars
    assert ast.literal_eval("'" + parser.dumps(parsed) + "'") == line



import functools
import json
import logging
import os
import time
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    List,
    Optional,
    Union,
    cast,
)
import warnings

import requests
from google.api_core import retry
from lbox.exceptions import ApiLimitError, NetworkError, ResourceNotFoundError
from tqdm import tqdm  # type: ignore

import labelbox
from labelbox import parser
from labelbox.orm import query
from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field, Relationship
from labelbox.schema.confidence_presence_checker import (
    LabelsConfidencePresenceChecker,
)
from labelbox.schema.enums import AnnotationImportState
from labelbox.schema.serialization import serialize_labels
from labelbox.utils import is_exactly_one_set

if TYPE_CHECKING:
    from labelbox.types import Label

NDJSON_MIME_TYPE = "application/x-ndjson"
ANNOTATION_PER_LABEL_LIMIT = 5000

logger = logging.getLogger(__name__)


class AnnotationImport(DbObject):
    name = Field.String("name")
    state = Field.Enum(AnnotationImportState, "state")
    input_file_url = Field.String("input_file_url")
    error_file_url = Field.String("error_file_url")
    status_file_url = Field.String("status_file_url")
    progress = Field.String("progress")

    created_by = Relationship.ToOne("User", False, "created_by")

    @property
    def inputs(self) -> List[Dict[str, Any]]:
        """
        Inputs for each individual annotation uploaded.
        This should match the ndjson annotations that you have uploaded.
        Returns:
            Uploaded ndjson.
        * This information will expire after 24 hours.
        """
        return self._fetch_remote_ndjson(self.input_file_url)

    @property
    def errors(self) -> List[Dict[str, Any]]:
        """
        Errors for each individual annotation uploaded. This is a subset of statuses

        Returns:
            List of dicts containing error messages. Empty list means there were no errors
            See `AnnotationImport.statuses` for more details.
        * This information will expire after 24 hours.
        """
        self.wait_till_done()
        return self._fetch_remote_ndjson(self.error_file_url)

    @property
    def statuses(self) -> List[Dict[str, Any]]:
        """
        Status for each individual annotation uploaded.

        Returns:
            A status for each annotation if the upload is done running.
            See below table for more details

        .. list-table::
            :widths: 15 150
            :header-rows: 1

            * - Field
              - Description
            * - uuid
              - Specifies the annotation for the status row.
            * - dataRow
              - JSON object containing the Labelbox data row ID for the annotation.
            * - status
              - Indicates SUCCESS or FAILURE.
            * - errors
              - An array of error messages included when status is FAILURE. Each error has a name, message and optional (key might not exist) additional_info.

        * This information will expire after 24 hours.
        """
        self.wait_till_done()
        return self._fetch_remote_ndjson(self.status_file_url)

    def wait_until_done(
        self, sleep_time_seconds: int = 10, show_progress: bool = False
    ) -> None:
        warnings.warn(
            "The method wait_until_done for AnnotationImport is deprecated and will be removed in the next major release. Use the wait_till_done method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.wait_till_done(sleep_time_seconds, show_progress)

    def wait_till_done(
        self, sleep_time_seconds: int = 10, show_progress: bool = False
    ) -> None:
        """Blocks import job until certain conditions are met.
        Blocks until the AnnotationImport.state changes either to
        `AnnotationImportState.FINISHED` or `AnnotationImportState.FAILED`,
        periodically refreshing object's state.
        Args:
            sleep_time_seconds (int): a time to block between subsequent API calls
            show_progress (bool): should show progress bar
        """
        pbar = (
            tqdm(
                total=100,
                bar_format="{n}% |{bar}| [{elapsed}, {rate_fmt}{postfix}]",
            )
            if show_progress
            else None
        )
        while self.state.value == AnnotationImportState.RUNNING.value:
            logger.info(f"Sleeping for {sleep_time_seconds} seconds...")
            time.sleep(sleep_time_seconds)
            self.__backoff_refresh()
            if self.progress and self.progress and pbar:
                pbar.update(int(self.progress.replace("%", "")) - pbar.n)

        if pbar:
            pbar.update(100 - pbar.n)
            pbar.close()

    @retry.Retry(
        predicate=retry.if_exception_type(
            ApiLimitError,
            TimeoutError,
            NetworkError,
        )
    )
    def __backoff_refresh(self) -> None:
        self.refresh()

    @functools.lru_cache()
    def _fetch_remote_ndjson(self, url: str) -> List[Dict[str, Any]]:
        """
        Fetches the remote ndjson file and caches the results.
        Args:
            url (str): Can be any url pointing to an ndjson file.
        Returns:
            ndjson as a list of dicts.
        """
        if self.state == AnnotationImportState.FAILED:
            raise ValueError("Import failed.")

        response = requests.get(url)
        response.raise_for_status()
        return parser.loads(response.text)

    @classmethod
    def _create_from_bytes(
        cls, client, variables, query_str, file_name, bytes_data
    ) -> Dict[str, Any]:
        operations = json.dumps({"variables": variables, "query": query_str})
        data = {
            "operations": operations,
            "map": (None, json.dumps({file_name: ["variables.file"]})),
        }
        file_data = (file_name, bytes_data, NDJSON_MIME_TYPE)
        files = {file_name: file_data}
        return client.execute(data=data, files=files)

    @classmethod
    def _get_ndjson_from_objects(
        cls,
        objects: Union[List[Dict[str, Any]], List["Label"]],
        object_name: str,
    ) -> BinaryIO:
        if not isinstance(objects, list):
            raise TypeError(
                f"{object_name} must be in a form of list. Found {type(objects)}"
            )

        objects = serialize_labels(objects)
        cls._validate_data_rows(objects)

        data_str = parser.dumps(objects)
        if not data_str:
            raise ValueError(f"{object_name} cannot be empty")

        return data_str.encode(
            "utf-8"
        )  # NOTICE this method returns bytes, NOT BinaryIO... should have done  io.BytesIO(...) but not going to change this at the moment since it works and fools mypy

    def refresh(self) -> None:
        """Synchronizes values of all fields with the database."""
        cls = type(self)
        res = cls.from_name(
            self.client, self.parent_id, self.name, as_json=True
        )
        self._set_field_values(res)

    @classmethod
    def _validate_data_rows(cls, objects: List[Dict[str, Any]]):
        """
        Validates annotations by checking 'dataRow' is provided
        and only one of 'id' or 'globalKey' is provided.

        Shows up to `max_num_errors` errors if invalidated, to prevent
        large number of error messages from being printed out
        """
        errors = []
        max_num_errors = 100
        labels_per_datarow: Dict[str, Dict[str, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        for object in objects:
            if "dataRow" not in object:
                errors.append(f"'dataRow' is missing in {object}")
                continue
            data_row_object = object["dataRow"]
            if not is_exactly_one_set(
                data_row_object.get("id"), data_row_object.get("globalKey")
            ):
                errors.append(
                    f"Must provide only one of 'id' or 'globalKey' for 'dataRow' in {object}"
                )
            else:
                data_row_id = data_row_object.get(
                    "globalKey"
                ) or data_row_object.get("id")
                name = object.get("name")
                if name:
                    labels_per_datarow[data_row_id][name] += 1
        for data_row_id, label_annotations in labels_per_datarow.items():
            for label_name, annotations in label_annotations.items():
                if annotations > ANNOTATION_PER_LABEL_LIMIT:
                    errors.append(
                        f"Row with id or global key {data_row_id} has {annotations} annotations for label {label_name}.\
                              Imports are limited to {ANNOTATION_PER_LABEL_LIMIT} annotations per data row."
                    )
        if errors:
            errors_length = len(errors)
            formatted_errors = "\n".join(errors[:max_num_errors])
            if errors_length > max_num_errors:
                logger.warning(
                    f"Found more than {max_num_errors} errors. Showing first {max_num_errors} error messages..."
                )
            raise ValueError(
                f"Error while validating annotations. Found {errors_length} annotations with errors. Errors:\n{formatted_errors}"
            )

    @classmethod
    def from_name(
        cls,
        client: "labelbox.Client",
        parent_id: str,
        name: str,
        as_json: bool = False,
    ):
        raise NotImplementedError("Inheriting class must override")

    @property
    def parent_id(self) -> str:
        raise NotImplementedError("Inheriting class must override")


class CreatableAnnotationImport(AnnotationImport):
    @classmethod
    def create(
        cls,
        client: "labelbox.Client",
        id: str,
        name: str,
        path: Optional[str] = None,
        url: Optional[str] = None,
        labels: Union[List[Dict[str, Any]], List["Label"]] = [],
    ) -> "AnnotationImport":
        if not is_exactly_one_set(url, labels, path):
            raise ValueError(
                "Must pass in a nonempty argument for one and only one of the following arguments: url, path, predictions"
            )
        if url:
            return cls.create_from_url(client, id, name, url)
        if path:
            return cls.create_from_file(client, id, name, path)
        return cls.create_from_objects(client, id, name, labels)

    @classmethod
    def create_from_url(
        cls, client: "labelbox.Client", id: str, name: str, url: str
    ) -> "AnnotationImport":
        raise NotImplementedError("Inheriting class must override")

    @classmethod
    def create_from_file(
        cls, client: "labelbox.Client", id: str, name: str, path: str
    ) -> "AnnotationImport":
        raise NotImplementedError("Inheriting class must override")

    @classmethod
    def create_from_objects(
        cls,
        client: "labelbox.Client",
        id: str,
        name: str,
        labels: Union[List[Dict[str, Any]], List["Label"]],
    ) -> "AnnotationImport":
        raise NotImplementedError("Inheriting class must override")


class MEAPredictionImport(CreatableAnnotationImport):
    model_run_id = Field.String("model_run_id")

    @property
    def parent_id(self) -> str:
        """
        Identifier for this import. Used to refresh the status
        """
        return self.model_run_id

    @classmethod
    def create_from_file(
        cls, client: "labelbox.Client", model_run_id: str, name: str, path: str
    ) -> "MEAPredictionImport":
        """
        Create an MEA prediction import job from a file of annotations

        Args:
            client: Labelbox Client for executing queries
            model_run_id: Model run to import labels into
            name: Name of the import job. Can be used to reference the task later
            path: Path to ndjson file containing annotations
        Returns:
            MEAPredictionImport
        """
        if os.path.exists(path):
            with open(path, "rb") as f:
                return cls._create_mea_import_from_bytes(
                    client, model_run_id, name, f, os.stat(path).st_size
                )
        else:
            raise ValueError(f"File {path} is not accessible")

    @classmethod
    def create_from_objects(
        cls,
        client: "labelbox.Client",
        model_run_id: str,
        name,
        predictions: Union[List[Dict[str, Any]], List["Label"]],
    ) -> "MEAPredictionImport":
        """
        Create an MEA prediction import job from an in memory dictionary

        Args:
            client: Labelbox Client for executing queries
            model_run_id: Model run to import labels into
            name: Name of the import job. Can be used to reference the task later
            predictions: List of prediction annotations
        Returns:
            MEAPredictionImport
        """
        data = cls._get_ndjson_from_objects(predictions, "annotations")

        return cls._create_mea_import_from_bytes(
            client, model_run_id, name, data, len(str(data))
        )

    @classmethod
    def create_from_url(
        cls, client: "labelbox.Client", model_run_id: str, name: str, url: str
    ) -> "MEAPredictionImport":
        """
        Create an MEA prediction import job from a url
        The url must point to a file containing prediction annotations.

        Args:
            client: Labelbox Client for executing queries
            model_run_id: Model run to import labels into
            name: Name of the import job. Can be used to reference the task later
            url: Url pointing to file to upload
        Returns:
            MEAPredictionImport
        """
        if requests.head(url):
            query_str = cls._get_url_mutation()
            return cls(
                client,
                client.execute(
                    query_str,
                    params={
                        "fileUrl": url,
                        "modelRunId": model_run_id,
                        "name": name,
                    },
                )["createModelErrorAnalysisPredictionImport"],
            )
        else:
            raise ValueError(f"Url {url} is not reachable")

    @classmethod
    def from_name(
        cls,
        client: "labelbox.Client",
        model_run_id: str,
        name: str,
        as_json: bool = False,
    ) -> "MEAPredictionImport":
        """
        Retrieves an MEA import job.

        Args:
            client: Labelbox Client for executing queries
            model_run_id: ID used for querying import jobs
            name: Name of the import job.
        Returns:
            MEAPredictionImport
        """
        query_str = """query getModelErrorAnalysisPredictionImportPyApi($modelRunId : ID!, $name: String!) {
            modelErrorAnalysisPredictionImport(
                where: {modelRunId: $modelRunId, name: $name}){
                    %s
                }}""" % query.results_query_part(cls)
        params = {
            "modelRunId": model_run_id,
            "name": name,
        }
        response = client.execute(query_str, params)
        if response is None:
            raise ResourceNotFoundError(MEAPredictionImport, params)
        response = response["modelErrorAnalysisPredictionImport"]
        if as_json:
            return response
        return cls(client, response)

    @classmethod
    def _get_url_mutation(cls) -> str:
        return """mutation createMEAPredictionImportByUrlPyApi($modelRunId : ID!, $name: String!, $fileUrl: String!) {
            createModelErrorAnalysisPredictionImport(data: {
                modelRunId: $modelRunId
                name: $name
                fileUrl: $fileUrl
            }) {%s}
        }""" % query.results_query_part(cls)

    @classmethod
    def _get_file_mutation(cls) -> str:
        return """mutation createMEAPredictionImportByFilePyApi($modelRunId : ID!, $name: String!, $file: Upload!, $contentLength: Int!) {
            createModelErrorAnalysisPredictionImport(data: {
                modelRunId: $modelRunId name: $name filePayload: { file: $file, contentLength: $contentLength}
        }) {%s}
        }""" % query.results_query_part(cls)

    @classmethod
    def _create_mea_import_from_bytes(
        cls,
        client: "labelbox.Client",
        model_run_id: str,
        name: str,
        bytes_data: BinaryIO,
        content_len: int,
    ) -> "MEAPredictionImport":
        file_name = f"{model_run_id}__{name}.ndjson"
        variables = {
            "file": None,
            "contentLength": content_len,
            "modelRunId": model_run_id,
            "name": name,
        }
        query_str = cls._get_file_mutation()
        res = cls._create_from_bytes(
            client,
            variables,
            query_str,
            file_name,
            bytes_data,
        )
        return cls(client, res["createModelErrorAnalysisPredictionImport"])


class MEAToMALPredictionImport(AnnotationImport):
    project = Relationship.ToOne("Project", cache=True)

    @property
    def parent_id(self) -> str:
        """
        Identifier for this import. Used to refresh the status
        """
        return self.project().uid

    @classmethod
    def create_for_model_run_data_rows(
        cls,
        client: "labelbox.Client",
        model_run_id: str,
        data_row_ids: List[str],
        project_id: str,
        name: str,
    ) -> "MEAToMALPredictionImport":
        """
        Create an MEA to MAL prediction import job from a list of data row ids of a specific model run

        Args:
            client: Labelbox Client for executing queries
            data_row_ids: A list of data row ids
            model_run_id: model run id
        Returns:
            MEAToMALPredictionImport
        """
        query_str = cls._get_model_run_data_rows_mutation()
        return cls(
            client,
            client.execute(
                query_str,
                params={
                    "dataRowIds": data_row_ids,
                    "modelRunId": model_run_id,
                    "projectId": project_id,
                    "name": name,
                },
            )["createMalPredictionImportForModelRunDataRows"],
        )

    @classmethod
    def from_name(
        cls,
        client: "labelbox.Client",
        project_id: str,
        name: str,
        as_json: bool = False,
    ) -> "MEAToMALPredictionImport":
        """
        Retrieves an MEA to MAL import job.

        Args:
            client: Labelbox Client for executing queries
            project_id:  ID used for querying import jobs
            name: Name of the import job.
        Returns:
            MALPredictionImport
        """
        query_str = """query getMEAToMALPredictionImportPyApi($projectId : ID!, $name: String!) {
            meaToMalPredictionImport(
                where: {projectId: $projectId, name: $name}){
                    %s
                }}""" % query.results_query_part(cls)
        params = {
            "projectId": project_id,
            "name": name,
        }
        response = client.execute(query_str, params)
        if response is None:
            raise ResourceNotFoundError(MALPredictionImport, params)
        response = response["meaToMalPredictionImport"]
        if as_json:
            return response
        return cls(client, response)

    @classmethod
    def _get_model_run_data_rows_mutation(cls) -> str:
        return """mutation createMalPredictionImportForModelRunDataRowsPyApi($dataRowIds: [ID!]!, $name: String!, $modelRunId: ID!, $projectId:ID!) {
            createMalPredictionImportForModelRunDataRows(data: {
                name: $name
                modelRunId: $modelRunId
                dataRowIds: $dataRowIds
                projectId: $projectId
            }) {%s}
        }""" % query.results_query_part(cls)


class MALPredictionImport(CreatableAnnotationImport):
    project = Relationship.ToOne("Project", cache=True)

    @property
    def parent_id(self) -> str:
        """
        Identifier for this import. Used to refresh the status
        """
        return self.project().uid

    def delete(self) -> None:
        """
        Deletes a MALPredictionImport job
        """

        query_string = """
            mutation deleteModelAssistedLabelingPredictionImportPyApi($id: ID!) {
                deleteModelAssistedLabelingPredictionImport(where: { id: $id }) {
                        id
                    }
                }
            """
        self.client.execute(query_string, {"id": self.uid})

    @classmethod
    def create_from_file(
        cls, client: "labelbox.Client", project_id: str, name: str, path: str
    ) -> "MALPredictionImport":
        """
        Create an MAL prediction import job from a file of annotations

        Args:
            client: Labelbox Client for executing queries
            project_id: Project to import labels into
            name: Name of the import job. Can be used to reference the task later
            path: Path to ndjson file containing annotations
        Returns:
            MALPredictionImport
        """
        if os.path.exists(path):
            with open(path, "rb") as f:
                return cls._create_mal_import_from_bytes(
                    client, project_id, name, f, os.stat(path).st_size
                )
        else:
            raise ValueError(f"File {path} is not accessible")

    @classmethod
    def create_from_objects(
        cls,
        client: "labelbox.Client",
        project_id: str,
        name: str,
        predictions: Union[List[Dict[str, Any]], List["Label"]],
    ) -> "MALPredictionImport":
        """
        Create an MAL prediction import job from an in memory dictionary

        Args:
            client: Labelbox Client for executing queries
            project_id: Project to import labels into
            name: Name of the import job. Can be used to reference the task later
            predictions: List of prediction annotations
        Returns:
            MALPredictionImport
        """

        data = cls._get_ndjson_from_objects(predictions, "annotations")
        if len(predictions) > 0 and isinstance(predictions[0], Dict):
            predictions_dicts = cast(List[Dict[str, Any]], predictions)
            has_confidence = LabelsConfidencePresenceChecker.check(
                predictions_dicts
            )
            if has_confidence:
                logger.warning("""
                Confidence scores are not supported in MAL Prediction Import.
                Corresponding confidence score values will be ignored.
                """)
        return cls._create_mal_import_from_bytes(
            client, project_id, name, data, len(str(data))
        )

    @classmethod
    def create_from_url(
        cls, client: "labelbox.Client", project_id: str, name: str, url: str
    ) -> "MALPredictionImport":
        """
        Create an MAL prediction import job from a url
        The url must point to a file containing prediction annotations.

        Args:
            client: Labelbox Client for executing queries
            project_id: Project to import labels into
            name: Name of the import job. Can be used to reference the task later
            url: Url pointing to file to upload
        Returns:
            MALPredictionImport
        """
        if requests.head(url):
            query_str = cls._get_url_mutation()
            return cls(
                client,
                client.execute(
                    query_str,
                    params={
                        "fileUrl": url,
                        "projectId": project_id,
                        "name": name,
                    },
                )["createModelAssistedLabelingPredictionImport"],
            )
        else:
            raise ValueError(f"Url {url} is not reachable")

    @classmethod
    def from_name(
        cls,
        client: "labelbox.Client",
        project_id: str,
        name: str,
        as_json: bool = False,
    ) -> "MALPredictionImport":
        """
        Retrieves an MAL import job.

        Args:
            client: Labelbox Client for executing queries
            project_id:  ID used for querying import jobs
            name: Name of the import job.
        Returns:
            MALPredictionImport
        """
        query_str = """query getModelAssistedLabelingPredictionImportPyApi($projectId : ID!, $name: String!) {
            modelAssistedLabelingPredictionImport(
                where: {projectId: $projectId, name: $name}){
                    %s
                }}""" % query.results_query_part(cls)
        params = {
            "projectId": project_id,
            "name": name,
        }
        response = client.execute(query_str, params)
        if response is None:
            raise ResourceNotFoundError(MALPredictionImport, params)
        response = response["modelAssistedLabelingPredictionImport"]
        if as_json:
            return response
        return cls(client, response)

    @classmethod
    def _get_url_mutation(cls) -> str:
        return """mutation createMALPredictionImportByUrlPyApi($projectId : ID!, $name: String!, $fileUrl: String!) {
            createModelAssistedLabelingPredictionImport(data: {
                projectId: $projectId
                name: $name
                fileUrl: $fileUrl
            }) {%s}
        }""" % query.results_query_part(cls)

    @classmethod
    def _get_file_mutation(cls) -> str:
        return """mutation createMALPredictionImportByFilePyApi($projectId : ID!, $name: String!, $file: Upload!, $contentLength: Int!) {
            createModelAssistedLabelingPredictionImport(data: {
                projectId: $projectId name: $name filePayload: { file: $file, contentLength: $contentLength}
        }) {%s}
        }""" % query.results_query_part(cls)

    @classmethod
    def _create_mal_import_from_bytes(
        cls,
        client: "labelbox.Client",
        project_id: str,
        name: str,
        bytes_data: BinaryIO,
        content_len: int,
    ) -> "MALPredictionImport":
        file_name = f"{project_id}__{name}.ndjson"
        variables = {
            "file": None,
            "contentLength": content_len,
            "projectId": project_id,
            "name": name,
        }
        query_str = cls._get_file_mutation()
        res = cls._create_from_bytes(
            client, variables, query_str, file_name, bytes_data
        )
        return cls(client, res["createModelAssistedLabelingPredictionImport"])


class LabelImport(CreatableAnnotationImport):
    project = Relationship.ToOne("Project", cache=True)

    @property
    def parent_id(self) -> str:
        """
        Identifier for this import. Used to refresh the status
        """
        return self.project().uid

    @classmethod
    def create_from_file(
        cls, client: "labelbox.Client", project_id: str, name: str, path: str
    ) -> "LabelImport":
        """
        Create a label import job from a file of annotations

        Args:
            client: Labelbox Client for executing queries
            project_id: Project to import labels into
            name: Name of the import job. Can be used to reference the task later
            path: Path to ndjson file containing annotations
        Returns:
            LabelImport
        """
        if os.path.exists(path):
            with open(path, "rb") as f:
                return cls._create_label_import_from_bytes(
                    client, project_id, name, f, os.stat(path).st_size
                )
        else:
            raise ValueError(f"File {path} is not accessible")

    @classmethod
    def create_from_objects(
        cls,
        client: "labelbox.Client",
        project_id: str,
        name: str,
        labels: Union[List[Dict[str, Any]], List["Label"]],
    ) -> "LabelImport":
        """
        Create a label import job from an in memory dictionary

        Args:
            client: Labelbox Client for executing queries
            project_id: Project to import labels into
            name: Name of the import job. Can be used to reference the task later
            labels: List of labels
        Returns:
            LabelImport
        """
        data = cls._get_ndjson_from_objects(labels, "labels")

        if len(labels) > 0 and isinstance(labels[0], Dict):
            label_dicts = cast(List[Dict[str, Any]], labels)
            has_confidence = LabelsConfidencePresenceChecker.check(label_dicts)
            if has_confidence:
                logger.warning("""
                Confidence scores are not supported in Label Import.
                Corresponding confidence score values will be ignored.
                """)
        return cls._create_label_import_from_bytes(
            client, project_id, name, data, len(str(data))
        )

    @classmethod
    def create_from_url(
        cls, client: "labelbox.Client", project_id: str, name: str, url: str
    ) -> "LabelImport":
        """
        Create a label annotation import job from a url
        The url must point to a file containing label annotations.

        Args:
            client: Labelbox Client for executing queries
            project_id: Project to import labels into
            name: Name of the import job. Can be used to reference the task later
            url: Url pointing to file to upload
        Returns:
            LabelImport
        """
        if requests.head(url):
            query_str = cls._get_url_mutation()
            return cls(
                client,
                client.execute(
                    query_str,
                    params={
                        "fileUrl": url,
                        "projectId": project_id,
                        "name": name,
                    },
                )["createLabelImport"],
            )
        else:
            raise ValueError(f"Url {url} is not reachable")

    @classmethod
    def from_name(
        cls,
        client: "labelbox.Client",
        project_id: str,
        name: str,
        as_json: bool = False,
    ) -> "LabelImport":
        """
        Retrieves an label import job.

        Args:
            client: Labelbox Client for executing queries
            project_id:  ID used for querying import jobs
            name: Name of the import job.
        Returns:
            LabelImport
        """
        query_str = """query getLabelImportPyApi($projectId : ID!, $name: String!) {
            labelImport(
                where: {projectId: $projectId, name: $name}){
                    %s
                }}""" % query.results_query_part(cls)
        params = {
            "projectId": project_id,
            "name": name,
        }
        response = client.execute(query_str, params)
        if response is None:
            raise ResourceNotFoundError(LabelImport, params)
        response = response["labelImport"]
        if as_json:
            return response
        return cls(client, response)

    @classmethod
    def _get_url_mutation(cls) -> str:
        return """mutation createLabelImportByUrlPyApi($projectId : ID!, $name: String!, $fileUrl: String!) {
            createLabelImport(data: {
                projectId: $projectId
                name: $name
                fileUrl: $fileUrl
            }) {%s}
        }""" % query.results_query_part(cls)

    @classmethod
    def _get_file_mutation(cls) -> str:
        return """mutation createLabelImportByFilePyApi($projectId : ID!, $name: String!, $file: Upload!, $contentLength: Int!) {
            createLabelImport(data: {
                projectId: $projectId name: $name filePayload: { file: $file, contentLength: $contentLength}
        }) {%s}
        }""" % query.results_query_part(cls)

    @classmethod
    def _create_label_import_from_bytes(
        cls,
        client: "labelbox.Client",
        project_id: str,
        name: str,
        bytes_data: BinaryIO,
        content_len: int,
    ) -> "LabelImport":
        file_name = f"{project_id}__{name}.ndjson"
        variables = {
            "file": None,
            "contentLength": content_len,
            "projectId": project_id,
            "name": name,
        }
        query_str = cls._get_file_mutation()
        res = cls._create_from_bytes(
            client, variables, query_str, file_name, bytes_data
        )
        return cls(client, res["createLabelImport"])



from pytest_cases import fixture_ref
from pytest_cases import parametrize, fixture_ref

from labelbox.data.metrics.confusion_matrix.confusion_matrix import (
    confusion_matrix_metric,
)


@parametrize(
    "tool_examples",
    [
        fixture_ref("polygon_pairs"),
        fixture_ref("rectangle_pairs"),
        fixture_ref("mask_pairs"),
        fixture_ref("line_pairs"),
        fixture_ref("point_pairs"),
        fixture_ref("ner_pairs"),
    ],
)
def test_overlapping_objects(tool_examples):
    for example in tool_examples:
        for include_subclasses, expected_attr_name in [
            [True, "expected"],
            [False, "expected_without_subclasses"],
        ]:
            score = confusion_matrix_metric(
                example.ground_truths,
                example.predictions,
                include_subclasses=include_subclasses,
            )

            if len(getattr(example, expected_attr_name)) == 0:
                assert len(score) == 0
            else:
                expected = [0, 0, 0, 0]
                for expected_values in getattr(
                    example, expected_attr_name
                ).values():
                    for idx in range(4):
                        expected[idx] += expected_values[idx]
                assert score[0].value == tuple(
                    expected
                ), f"{example.predictions},{example.ground_truths}"


@parametrize(
    "tool_examples",
    [fixture_ref("checklist_pairs"), fixture_ref("radio_pairs")],
)
def test_overlapping_classifications(tool_examples):
    for example in tool_examples:
        score = confusion_matrix_metric(
            example.ground_truths, example.predictions
        )
        if len(example.expected) == 0:
            assert len(score) == 0
        else:
            expected = [0, 0, 0, 0]
            for expected_values in example.expected.values():
                for idx in range(4):
                    expected[idx] += expected_values[idx]
            assert score[0].value == tuple(
                expected
            ), f"{example.predictions},{example.ground_truths}"


def test_partial_overlap(pair_iou_thresholds):
    for example in pair_iou_thresholds:
        for iou in example.expected.keys():
            score = confusion_matrix_metric(
                example.predictions, example.ground_truths, iou=iou
            )
            assert score[0].value == tuple(
                example.expected[iou]
            ), f"{example.predictions},{example.ground_truths}"



import datetime
import re

from dateutil.tz import tzoffset
from dateutil.parser import isoparse as dateutil_parse
from dateutil.utils import default_tzinfo

from urllib.parse import urlparse
from pydantic import (
    BaseModel,
    ConfigDict,
    model_serializer,
)
from pydantic.alias_generators import to_camel

UPPERCASE_COMPONENTS = ["uri", "rgb"]
ISO_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
DFLT_TZ = tzoffset("UTC", 0000)


def _convert(s, sep, title):
    components = re.findall(r"[A-Z][a-z0-9]*|[a-z][a-z0-9]*", s)
    components = list(map(str.lower, filter(None, components)))
    for i in range(len(components)):
        if components[i] in UPPERCASE_COMPONENTS:
            components[i] = components[i].upper()
        elif title(i):
            components[i] = components[i][0].upper() + components[i][1:]
    return sep.join(components)


def camel_case(s):
    """Converts a string in [snake|camel|title]case to camelCase."""
    return _convert(s, "", lambda i: i > 0)


def title_case(s):
    """Converts a string in [snake|camel|title]case to TitleCase."""
    return _convert(s, "", lambda i: True)


def snake_case(s):
    """Converts a string in [snake|camel|title]case to snake_case."""
    return _convert(s, "_", lambda i: False)


def sentence_case(s: str) -> str:
    """Converts a string in [snake|camel|title]case to Sentence case."""
    # Replace underscores with spaces and convert to lower case
    sentence_str = s.replace("_", " ").lower()
    # Capitalize the first letter of each word
    sentence_str = sentence_str.capitalize()
    return sentence_str


def is_exactly_one_set(*args):
    return sum([bool(arg) for arg in args]) == 1


def is_valid_uri(uri):
    try:
        result = urlparse(uri)
        return all([result.scheme, result.netloc])
    except:
        return False


class _CamelCaseMixin(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        alias_generator=to_camel,
        populate_by_name=True,
    )


class _NoCoercionMixin:
    """
    When using Unions in type annotations, pydantic will try to coerce the type
    of the object to the type of the first Union member. Which results in
    uninteded behavior.

    This mixin uses a class_name discriminator field to prevent pydantic from
    corecing the type of the object. Add a class_name field to the class you
    want to discrimniate and use this mixin class to remove the discriminator
    when serializing the object.

    Example:
        class GenericDataRowData(BaseData, _NoCoercionMixin):
            class_name: Literal["GenericDataRowData"] = "GenericDataRowData"

    """

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        res.pop("class_name")
        return res


def format_iso_datetime(dt: datetime.datetime) -> str:
    """
    Formats a datetime object into the format: 2011-11-04T00:05:23Z
    Note that datetime.isoformat() outputs 2011-11-04T00:05:23+00:00
    """
    return dt.astimezone(datetime.timezone.utc).strftime(ISO_DATETIME_FORMAT)


def format_iso_from_string(date_string: str) -> datetime.datetime:
    """
    Converts a string even if offset is missing: 2011-11-04T00:05:23Z or 2011-11-04T00:05:23+00:00 or 2011-11-04T00:05:23
    to a datetime object.
    For missing offsets, the default offset is UTC.
    """
    # return datetime.datetime.fromisoformat(date_string)
    return default_tzinfo(dateutil_parse(date_string), DFLT_TZ)



from unittest.mock import MagicMock

import pytest

from labelbox.schema.export_filters import build_filters


def test_ids_filter():
    client = MagicMock()
    filters = {"data_row_ids": ["id1", "id2"], "batch_ids": ["b1", "b2"]}
    assert build_filters(client, filters) == [
        {
            "ids": ["id1", "id2"],
            "operator": "is",
            "type": "data_row_id",
        },
        {
            "ids": ["b1", "b2"],
            "operator": "is",
            "type": "batch",
        },
    ]


def test_ids_empty_filter():
    client = MagicMock()
    filters = {"data_row_ids": [], "batch_ids": ["b1", "b2"]}
    with pytest.raises(
        ValueError, match="data_row_id filter expects a non-empty list."
    ):
        build_filters(client, filters)


def test_global_keys_filter():
    client = MagicMock()
    filters = {"global_keys": ["id1", "id2"]}
    assert build_filters(client, filters) == [
        {
            "ids": ["id1", "id2"],
            "operator": "is",
            "type": "global_key",
        }
    ]


def test_validations():
    client = MagicMock()
    filters = {
        "global_keys": ["id1", "id2"],
        "data_row_ids": ["id1", "id2"],
    }
    with pytest.raises(
        ValueError,
        match="data_rows and global_keys cannot both be present in export filters",
    ):
        build_filters(client, filters)



import pytest
import cv2

from labelbox.data.annotation_types import Point
from pydantic import ValidationError


def test_point():
    with pytest.raises(ValidationError):
        line = Point()

    with pytest.raises(TypeError):
        line = Point([0, 1])

    point = Point(x=0, y=1)
    expected = {"coordinates": [0, 1], "type": "Point"}
    assert point.geometry == expected
    expected["coordinates"] = tuple(expected["coordinates"])
    assert point.shapely.__geo_interface__ == expected

    raster = point.draw(height=32, width=32, thickness=1)
    assert (cv2.imread("tests/data/assets/point.png") == raster).all()



from typing import List, Optional, Union, Tuple

import geojson
import numpy as np
import cv2

from shapely.geometry import LineString as SLineString

from .point import Point
from .geometry import Geometry

from pydantic import field_validator


class Line(Geometry):
    """Line annotation

    Args:
        points (List[Point]): A list of `Point` geometries

    >>> Line(points = [Point(x=3,y=4), Point(x=3,y=5)])

    """

    points: List[Point]

    @property
    def geometry(self) -> geojson.MultiLineString:
        return geojson.MultiLineString(
            [[[point.x, point.y] for point in self.points]]
        )

    @classmethod
    def from_shapely(cls, shapely_obj: SLineString) -> "Line":
        """Transforms a shapely object."""
        if not isinstance(shapely_obj, SLineString):
            raise TypeError(
                f"Expected Shapely Line. Got {shapely_obj.geom_type}"
            )

        obj_coords = shapely_obj.__geo_interface__["coordinates"]
        return Line(
            points=[Point(x=coords[0], y=coords[1]) for coords in obj_coords]
        )

    def draw(
        self,
        height: Optional[int] = None,
        width: Optional[int] = None,
        canvas: Optional[np.ndarray] = None,
        color: Union[int, Tuple[int, int, int]] = (255, 255, 255),
        thickness: int = 1,
    ) -> np.ndarray:
        """
        Draw the line onto a 3d mask
        Args:
            height (int): height of the mask
            width (int): width of the mask
            thickness (int): How thick to draw the line
            color (int): color for the line.
                  RGB values by default but if a 2D canvas is provided this can set this to an int.
            canvas (np.ndarry): Canvas for drawing line on.
        Returns:
            numpy array representing the mask with the line drawn on it.
        """
        canvas = self.get_or_create_canvas(height, width, canvas)
        pts = np.array(self.geometry["coordinates"]).astype(np.int32)
        return cv2.polylines(
            canvas, pts, False, color=color, thickness=thickness
        )

    @field_validator("points")
    def is_geom_valid(cls, points):
        if len(points) < 2:
            raise ValueError(
                f"A line must have at least 2 points to be valid. Found {points}"
            )

        return points



import json
import logging
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union
import warnings

import requests
from lbox.exceptions import ResourceNotFoundError

from labelbox import parser
from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Entity, Field, Relationship
from labelbox.pagination import PaginatedCollection
from labelbox.schema.internal.datarow_upload_constants import (
    DOWNLOAD_RESULT_PAGE_SIZE,
)

if TYPE_CHECKING:
    from labelbox import User

    def lru_cache() -> Callable[..., Callable[..., Dict[str, Any]]]:
        pass
else:
    from functools import lru_cache

logger = logging.getLogger(__name__)


class Task(DbObject):
    """Represents a server-side process that might take a longer time to process.
    Allows the Task state to be updated and checked on the client side.

    Attributes:
        updated_at (datetime)
        created_at (datetime)
        name (str)
        status (str)
        completion_percentage (float)

        created_by (Relationship): `ToOne` relationship to User
        organization (Relationship): `ToOne` relationship to Organization
    """

    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    name = Field.String("name")
    status = Field.String("status")
    completion_percentage = Field.Float("completion_percentage")
    result_url = Field.String("result_url", "result")
    errors_url = Field.String("errors_url", "errors")
    type = Field.String("type")
    metadata = Field.Json("metadata")
    _user: Optional["User"] = None

    # Relationships
    created_by = Relationship.ToOne("User", False, "created_by")
    organization = Relationship.ToOne("Organization")

    def __eq__(self, task):
        return (
            isinstance(task, Task)
            and task.uid == self.uid
            and task.type == self.type
        )

    def __hash__(self):
        return hash(self.uid)

    # Import and upsert have several instances of special casing
    def is_creation_task(self) -> bool:
        return self.name == "JSON Import" or self.type == "adv-upsert-data-rows"

    def refresh(self) -> None:
        """Refreshes Task data from the server."""
        assert self._user is not None
        tasks = list(self._user.created_tasks(where=Task.uid == self.uid))
        if len(tasks) != 1:
            raise ResourceNotFoundError(Task, self.uid)
        for field in self.fields():
            setattr(self, field.name, getattr(tasks[0], field.name))
            if self.is_creation_task():
                self.errors_url = self.result_url

    def has_errors(self) -> bool:
        if self.type == "export-data-rows":
            # self.errors fetches the error content.
            # This first condition prevents us from downloading the content for v2 exports
            return bool(self.errors_url or self.errors)
        if self.is_creation_task():
            return bool(self.failed_data_rows)
        return self.status == "FAILED"

    def wait_until_done(
        self, timeout_seconds: float = 300.0, check_frequency: float = 2.0
    ) -> None:
        warnings.warn(
            "The method wait_until_done for Task is deprecated and will be removed in the next major release. Use the wait_till_done method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.wait_till_done(timeout_seconds, check_frequency)

    def wait_till_done(
        self, timeout_seconds: float = 300.0, check_frequency: float = 2.0
    ) -> None:
        """Waits until the task is completed. Periodically queries the server
        to update the task attributes.

        Args:
            timeout_seconds (float): Maximum time this method can block, in seconds. Defaults to five minutes.
            check_frequency (float): Frequency of queries to server to update the task attributes, in seconds. Defaults to two seconds. Minimal value is two seconds.
        """
        if check_frequency < 2.0:
            raise ValueError(
                "Expected check frequency to be two seconds or more"
            )
        while timeout_seconds > 0:
            if self.status != "IN_PROGRESS":
                if self.has_errors():
                    logger.warning(
                        "There are errors present. Please look at `task.errors` for more details"
                    )
                return
            logger.debug(
                "Task.wait_till_done sleeping for %d seconds", check_frequency
            )
            time.sleep(check_frequency)
            timeout_seconds -= check_frequency
            self.refresh()

    @property
    def errors(self) -> Optional[Dict[str, Any]]:
        """Fetch the error associated with an import task."""
        if self.is_creation_task():
            if self.status == "FAILED":
                result = self._fetch_remote_json()
                return result["error"]
            elif self.status == "COMPLETE":
                return self.failed_data_rows
        elif self.type == "export-data-rows":
            return self._fetch_remote_json(remote_json_field="errors_url")
        elif (
            self.type == "add-data-rows-to-batch"
            or self.type == "send-to-task-queue"
            or self.type == "send-to-annotate"
        ):
            if self.status == "FAILED":
                # for these tasks, the error is embedded in the result itself
                return json.loads(self.result_url)
        return None

    @property
    def result(self) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """Fetch the result for an import task."""
        if self.status == "FAILED":
            raise ValueError(f"Job failed. Errors : {self.errors}")
        else:
            result = self._fetch_remote_json()
            if self.type == "export-data-rows":
                return result

            return [
                {
                    "id": data_row["id"],
                    "external_id": data_row.get("externalId"),
                    "row_data": data_row["rowData"],
                    "global_key": data_row.get("globalKey"),
                }
                for data_row in result["createdDataRows"]
            ]

    @property
    def failed_data_rows(self) -> Optional[Dict[str, Any]]:
        """Fetch data rows which failed to be created for an import task."""
        result = self._fetch_remote_json()
        if len(result.get("errors", [])) > 0:
            return result["errors"]
        else:
            return None

    @property
    def created_data_rows(self) -> Optional[Dict[str, Any]]:
        """Fetch data rows which successfully created for an import task."""
        result = self._fetch_remote_json()
        if len(result.get("createdDataRows", [])) > 0:
            return result["createdDataRows"]
        else:
            return None

    @lru_cache()
    def _fetch_remote_json(
        self, remote_json_field: Optional[str] = None
    ) -> Dict[str, Any]:
        """Function for fetching and caching the result data."""

        def download_result(remote_json_field: Optional[str], format: str):
            url = getattr(self, remote_json_field or "result_url")

            if url is None:
                return None

            response = requests.get(url)
            response.raise_for_status()
            if format == "json":
                return response.json()
            elif format == "ndjson":
                return parser.loads(response.text)
            else:
                raise ValueError(
                    "Expected the result format to be either `ndjson` or `json`."
                )

        if self.is_creation_task():
            format = "json"
        elif self.type == "export-data-rows":
            format = "ndjson"
        else:
            raise ValueError(
                "Task result is only supported for `JSON Import` and `export` tasks."
                " Download task.result_url manually to access the result for other tasks."
            )

        if self.status != "IN_PROGRESS":
            return download_result(remote_json_field, format)
        else:
            self.wait_till_done(timeout_seconds=600)
            if self.status == "IN_PROGRESS":
                raise ValueError(
                    "Job status still in `IN_PROGRESS`. The result is not available. Call task.wait_till_done() with a larger timeout or contact support."
                )
            return download_result(remote_json_field, format)

    @staticmethod
    def get_task(client, task_id):
        user: User = client.get_user()
        tasks: List[Task] = list(
            user.created_tasks(where=Entity.Task.uid == task_id)
        )
        # Cache user in a private variable as the relationship can't be
        # resolved due to server-side limitations (see Task.created_by)
        # for more info.
        if len(tasks) != 1:
            raise ResourceNotFoundError(Entity.Task, {task_id: task_id})
        task: Task = tasks[0]
        task._user = user
        return task


class DataUpsertTask(Task):
    """
    Task class for data row upsert operations
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._user = None

    @property
    def result(self) -> Optional[List[Dict[str, Any]]]:  # type: ignore
        """
        Fetches all results.
        Note, for large uploads (>150K data rows), it could take multiple minutes to complete
        """
        if self.status == "FAILED":
            raise ValueError(f"Job failed. Errors : {self.errors}")
        return self._results_as_list()

    @property
    def errors(self) -> Optional[List[Dict[str, Any]]]:  # type: ignore
        """
        Fetches all errors.
        Note, for large uploads / large number of errors (>150K), it could take multiple minutes to complete
        """
        return self._errors_as_list()

    @property
    def created_data_rows(  # type: ignore
        self,
    ) -> Optional[List[Dict[str, Any]]]:
        return self.result

    @property
    def failed_data_rows(  # type: ignore
        self,
    ) -> Optional[List[Dict[str, Any]]]:
        return self.errors

    def _download_results_paginated(self) -> PaginatedCollection:
        page_size = DOWNLOAD_RESULT_PAGE_SIZE
        from_cursor = None

        query_str = """query SuccessesfulDataRowImportsPyApi($taskId: ID!, $first: Int, $from: String)  {
                    successesfulDataRowImports(data: { taskId: $taskId, first: $first, from: $from})
                        {
                            nodes { 
                                id
                                externalId
                                globalKey
                                rowData
                                }
                            after
                            total
                        }
                    }
                """

        params = {
            "taskId": self.uid,
            "first": page_size,
            "from": from_cursor,
        }

        return PaginatedCollection(
            client=self.client,
            query=query_str,
            params=params,
            dereferencing=["successesfulDataRowImports", "nodes"],
            obj_class=lambda _, data_row: {
                "id": data_row.get("id"),
                "external_id": data_row.get("externalId"),
                "row_data": data_row.get("rowData"),
                "global_key": data_row.get("globalKey"),
            },
            cursor_path=["successesfulDataRowImports", "after"],
        )

    def _download_errors_paginated(self) -> PaginatedCollection:
        page_size = DOWNLOAD_RESULT_PAGE_SIZE  # hardcode to avoid overloading the server
        from_cursor = None

        query_str = """query FailedDataRowImportsPyApi($taskId: ID!, $first: Int, $from: String)  {
                    failedDataRowImports(data: { taskId: $taskId, first: $first, from: $from})
                        {
                            after
                            total
                            results {
                                message
                                spec {
                                    externalId
                                    globalKey
                                    rowData
                                        metadata {
                                            schemaId
                                            value
                                            name
                                        }
                                        attachments {
                                            type
                                            value
                                            name
                                        }                                    
                                }
                            }
                        }
                    }
                """

        params = {
            "taskId": self.uid,
            "first": page_size,
            "from": from_cursor,
        }

        def convert_errors_to_legacy_format(client, data_row):
            spec = data_row.get("spec", {})
            return {
                "message": data_row.get("message"),
                "failedDataRows": [
                    {
                        "externalId": spec.get("externalId"),
                        "rowData": spec.get("rowData"),
                        "globalKey": spec.get("globalKey"),
                        "metadata": spec.get("metadata", []),
                        "attachments": spec.get("attachments", []),
                    }
                ],
            }

        return PaginatedCollection(
            client=self.client,
            query=query_str,
            params=params,
            dereferencing=["failedDataRowImports", "results"],
            obj_class=convert_errors_to_legacy_format,
            cursor_path=["failedDataRowImports", "after"],
        )

    def _results_as_list(self) -> Optional[List[Dict[str, Any]]]:
        total_downloaded = 0
        results = []
        data = self._download_results_paginated()

        for row in data:
            results.append(row)
            total_downloaded += 1

        if len(results) == 0:
            return None

        return results

    def _errors_as_list(self) -> Optional[List[Dict[str, Any]]]:
        total_downloaded = 0
        errors = []
        data = self._download_errors_paginated()

        for row in data:
            errors.append(row)
            total_downloaded += 1

        if len(errors) == 0:
            return None

        return errors



import pytest
from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.annotation_types.label import Label


def test_generic_data_type():
    data = {
        "global_key": "https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr",
    }
    label = Label(data=data)
    data = label.data
    assert isinstance(data, GenericDataRowData)
    assert (
        data.global_key
        == "https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr"
    )


def test_generic_data_type_validations():
    data = {
        "row_data": "https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr",
    }
    with pytest.raises(ValueError, match="Exactly one of"):
        Label(data=data)

    data = {
        "uid": "abcd",
        "global_key": "https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr",
    }
    with pytest.raises(ValueError, match="Only one of"):
        Label(data=data)


def test_generic_data_row():
    data = {
        "global_key": "https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr",
    }
    label = Label(data=GenericDataRowData(**data))
    data = label.data
    assert isinstance(data, GenericDataRowData)
    assert (
        data.global_key
        == "https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr"
    )



from enum import Enum
from typing import Optional, Union

from labelbox.schema.media_type import MediaType


class OntologyKind(Enum):
    """
    OntologyKind is an enum that represents the different types of ontologies
    """

    ModelEvaluation = "MODEL_EVALUATION"
    ResponseCreation = "RESPONSE_CREATION"
    Missing = None

    @classmethod
    def is_supported(cls, value):
        return isinstance(value, cls)

    @classmethod
    def get_ontology_kind_validation_error(cls, ontology_kind):
        return TypeError(
            f"{ontology_kind}: is not a valid ontology kind. Use"
            f" any of {OntologyKind.__members__.items()}"
            " from OntologyKind."
        )

    @staticmethod
    def evaluate_ontology_kind_with_media_type(
        ontology_kind, media_type: Optional[MediaType]
    ) -> Union[MediaType, None]:
        ontology_to_media = {
            OntologyKind.ModelEvaluation: (
                MediaType.Conversational,
                "For chat evaluation, media_type must be Conversational.",
            ),
            OntologyKind.ResponseCreation: (
                MediaType.Text,
                "For response creation, media_type must be Text.",
            ),
        }

        if ontology_kind in ontology_to_media:
            expected_media_type, error_message = ontology_to_media[
                ontology_kind
            ]

            if media_type is None or media_type == expected_media_type:
                media_type = expected_media_type
            else:
                raise ValueError(error_message)

        return media_type


class EditorTaskType(str, Enum):
    ModelChatEvaluation = "MODEL_CHAT_EVALUATION"
    ResponseCreation = "RESPONSE_CREATION"
    OfflineModelChatEvaluation = "OFFLINE_MODEL_CHAT_EVALUATION"
    Missing = None

    @classmethod
    def is_supported(cls, value):
        return isinstance(value, cls)

    @classmethod
    def _missing_(cls, value) -> "EditorTaskType":
        """Handle missing null new task types
        Handle upper case names for compatibility with
        the GraphQL"""

        if value is None:
            return cls.Missing

        for name, member in cls.__members__.items():
            if value == name.upper():
                return member

        return cls.Missing


class EditorTaskTypeMapper:
    @staticmethod
    def to_editor_task_type(
        ontology_kind: OntologyKind, media_type: MediaType
    ) -> EditorTaskType:
        if (
            ontology_kind
            and OntologyKind.is_supported(ontology_kind)
            and media_type
            and MediaType.is_supported(media_type)
        ):
            editor_task_type = EditorTaskTypeMapper.map_to_editor_task_type(
                ontology_kind, media_type
            )
        else:
            editor_task_type = EditorTaskType.Missing

        return editor_task_type

    @staticmethod
    def map_to_editor_task_type(
        onotology_kind: OntologyKind, media_type: MediaType
    ) -> EditorTaskType:
        if (
            onotology_kind == OntologyKind.ModelEvaluation
            and media_type == MediaType.Conversational
        ):
            return EditorTaskType.ModelChatEvaluation
        elif (
            onotology_kind == OntologyKind.ResponseCreation
            and media_type == MediaType.Text
        ):
            return EditorTaskType.ResponseCreation
        else:
            return EditorTaskType.Missing


class UploadType(Enum):
    Auto = ("AUTO",)
    Manual = ("MANUAL",)
    Missing = None

    @classmethod
    def is_supported(cls, value):
        return isinstance(value, cls)

    @classmethod
    def _missing_(cls, value: object) -> "UploadType":
        if value is None:
            return cls.Missing

        for name, member in cls.__members__.items():
            if value == name.upper():
                return member

        return cls.Missing



from collections import defaultdict
from unittest.mock import MagicMock

import pytest
from lbox.exceptions import (
    MalformedQueryException,
    ResourceConflict,
    ResourceCreationError,
    ResourceNotFoundError,
    UnprocessableEntityError,
)

from labelbox import Client
from labelbox.schema.media_type import MediaType
from labelbox.schema.ontology_kind import EditorTaskType
from labelbox.schema.project import Project
from labelbox.schema.user import User
from labelbox.schema.user_group import UserGroup, UserGroupColor


@pytest.fixture
def group_user():
    user_values = defaultdict(lambda: None)
    user_values["id"] = "user_id"
    user_values["email"] = "test@example.com"
    return User(MagicMock(Client), user_values)


@pytest.fixture
def group_project():
    project_values = defaultdict(lambda: None)
    project_values["id"] = "project_id"
    project_values["name"] = "Test Project"
    project_values["editorTaskType"] = EditorTaskType.Missing.value
    project_values["mediaType"] = MediaType.Image.value
    return Project(MagicMock(Client), project_values)


class TestUserGroupColor:
    def test_user_group_color_values(self):
        assert UserGroupColor.BLUE.value == "9EC5FF"
        assert UserGroupColor.PURPLE.value == "CEB8FF"
        assert UserGroupColor.ORANGE.value == "FFB35F"
        assert UserGroupColor.CYAN.value == "4ED2F9"
        assert UserGroupColor.PINK.value == "FFAEA9"
        assert UserGroupColor.LIGHT_PINK.value == "FFA9D5"
        assert UserGroupColor.GREEN.value == "3FDC9A"
        assert UserGroupColor.YELLOW.value == "E7BF00"
        assert UserGroupColor.GRAY.value == "B8C4D3"


class TestUserGroup:
    def setup_method(self):
        self.client = MagicMock(Client)
        self.client.enable_experimental = True
        self.group = UserGroup(client=self.client)

    def test_constructor(self):
        group = UserGroup(self.client)

        assert group.id == ""
        assert group.name == ""
        assert group.color is UserGroupColor.BLUE
        assert len(group.projects) == 0
        assert len(group.users) == 0

    def test_update_with_exception_name(self):
        group = self.group
        group.id = ""

        with pytest.raises(ValueError):
            group.get()

    def test_get(self):
        projects = [
            {"id": "project_id_1", "name": "project_1"},
            {"id": "project_id_2", "name": "project_2"},
        ]
        group_members = [
            {"id": "user_id_1", "email": "email_1"},
            {"id": "user_id_2", "email": "email_2"},
        ]
        self.client.execute.return_value = {
            "userGroup": {
                "id": "group_id",
                "name": "Test Group",
                "color": "4ED2F9",
                "projects": {"nodes": projects},
                "members": {"nodes": group_members},
            }
        }
        group = UserGroup(self.client)
        assert group.id == ""
        assert group.name == ""
        assert group.color is UserGroupColor.BLUE
        assert len(group.projects) == 0
        assert len(group.users) == 0

        group.id = "group_id"
        group.get()

        assert group.id == "group_id"
        assert group.name == "Test Group"
        assert group.color is UserGroupColor.CYAN
        assert len(group.projects) == 2
        assert len(group.users) == 2

    def test_get_value_error(self):
        self.client.execute.return_value = None
        group = UserGroup(self.client)
        group.name = "Test Group"

        with pytest.raises(ValueError):
            group.get()

    def test_update(self, group_user, group_project):
        group = self.group
        group.id = "group_id"
        group.name = "Test Group"
        group.color = UserGroupColor.BLUE
        group.users = {group_user}
        group.projects = {group_project}

        updated_group = group.update()

        execute = self.client.execute.call_args[0]

        assert "UpdateUserGroupPyApi" in execute[0]
        assert execute[1]["id"] == "group_id"
        assert execute[1]["name"] == "Test Group"
        assert execute[1]["color"] == UserGroupColor.BLUE.value
        assert len(execute[1]["userIds"]) == 1
        assert list(execute[1]["userIds"])[0] == group_user.uid
        assert len(execute[1]["projectIds"]) == 1
        assert list(execute[1]["projectIds"])[0] == group_project.uid

        assert updated_group.id == "group_id"
        assert updated_group.name == "Test Group"
        assert updated_group.color == UserGroupColor.BLUE
        assert len(updated_group.users) == 1
        assert list(updated_group.users)[0].uid == group_user.uid
        assert len(updated_group.projects) == 1
        assert list(updated_group.projects)[0].uid == group_project.uid

    def test_update_resource_error_input_bad(self):
        self.client.execute.side_effect = MalformedQueryException("Error")
        group = UserGroup(self.client)
        group.name = "Test Group"
        group.id = "group_id"

        with pytest.raises(UnprocessableEntityError):
            group.update()

    def test_update_resource_error_unknown_id(self):
        self.client.execute.return_value = None
        group = UserGroup(self.client)
        group.name = "Test Group"
        group.id = "group_id"

        with pytest.raises(ResourceNotFoundError) as e:
            group.update()

    def test_update_with_exception_name(self):
        group = self.group
        group.name = ""

        with pytest.raises(UnprocessableEntityError):
            group.update()

    def test_update_with_exception_name(self):
        group = self.group
        group.id = ""

        with pytest.raises(ValueError):
            group.update()

    def test_create_with_exception_id(self):
        group = self.group
        group.id = "group_id"

        with pytest.raises(ResourceCreationError):
            group.create()

    def test_create_with_exception_name(self):
        group = self.group
        group.name = ""

        with pytest.raises(ValueError):
            group.create()

    def test_create(self, group_user, group_project):
        group = self.group
        group.name = "New Group"
        group.color = UserGroupColor.PINK
        group.users = {group_user}
        group.projects = {group_project}

        self.client.execute.return_value = {
            "createUserGroup": {"group": {"id": "group_id"}}
        }
        created_group = group.create()
        execute = self.client.execute.call_args[0]

        assert "CreateUserGroupPyApi" in execute[0]
        assert execute[1]["name"] == "New Group"
        assert execute[1]["color"] == UserGroupColor.PINK.value
        assert len(execute[1]["userIds"]) == 1
        assert list(execute[1]["userIds"])[0] == "user_id"
        assert len(execute[1]["projectIds"]) == 1
        assert list(execute[1]["projectIds"])[0] == "project_id"
        assert created_group.id is not None
        assert created_group.id == "group_id"
        assert created_group.name == "New Group"
        assert created_group.color == UserGroupColor.PINK
        assert len(created_group.users) == 1
        assert list(created_group.users)[0].uid == "user_id"
        assert len(created_group.projects) == 1
        assert list(created_group.projects)[0].uid == "project_id"

    def test_create_resource_creation_error(self):
        self.client.execute.side_effect = ResourceConflict("Error")
        group = UserGroup(self.client)
        group.name = "Test Group"

        with pytest.raises(ResourceCreationError):
            group.create()

    def test_delete(self):
        group = self.group
        group.id = "group_id"

        self.client.execute.return_value = {
            "deleteUserGroup": {"success": True}
        }
        deleted = group.delete()
        execute = self.client.execute.call_args[0]

        assert "DeleteUserGroupPyApi" in execute[0]
        assert execute[1]["id"] == "group_id"
        assert deleted is True

    def test_delete_resource_not_found_error(self):
        self.client.execute.return_value = None
        group = UserGroup(self.client)
        group.id = "group_id"

        with pytest.raises(ResourceNotFoundError):
            group.delete()

    def test_delete_no_id(self):
        group = UserGroup(self.client)
        group.id = None

        with pytest.raises(ValueError):
            group.delete()

    def test_user_groups_empty(self):
        self.client.execute.return_value = {"userGroups": None}

        user_groups = list(UserGroup(self.client).get_user_groups())

        assert len(user_groups) == 0

    def test_user_groups(self):
        self.client.execute.return_value = {
            "userGroups": {
                "nextCursor": None,
                "nodes": [
                    {
                        "id": "group_id_1",
                        "name": "Group 1",
                        "color": "9EC5FF",
                        "projects": {
                            "nodes": [
                                {"id": "project_id_1", "name": "Project 1"},
                                {"id": "project_id_2", "name": "Project 2"},
                            ]
                        },
                        "members": {
                            "nodes": [
                                {
                                    "id": "user_id_1",
                                    "email": "user1@example.com",
                                },
                                {
                                    "id": "user_id_2",
                                    "email": "user2@example.com",
                                },
                            ]
                        },
                    },
                    {
                        "id": "group_id_2",
                        "name": "Group 2",
                        "color": "9EC5FF",
                        "projects": {
                            "nodes": [
                                {"id": "project_id_3", "name": "Project 3"},
                                {"id": "project_id_4", "name": "Project 4"},
                            ]
                        },
                        "members": {
                            "nodes": [
                                {
                                    "id": "user_id_3",
                                    "email": "user3@example.com",
                                },
                                {
                                    "id": "user_id_4",
                                    "email": "user4@example.com",
                                },
                            ]
                        },
                    },
                    {
                        "id": "group_id_3",
                        "name": "Group 3",
                        "color": "9EC5FF",
                        "projects": {
                            "nodes": [
                                {"id": "project_id_5", "name": "Project 5"},
                                {"id": "project_id_6", "name": "Project 6"},
                            ]
                        },
                        "members": {
                            "nodes": [
                                {
                                    "id": "user_id_5",
                                    "email": "user5@example.com",
                                },
                                {
                                    "id": "user_id_6",
                                    "email": "user6@example.com",
                                },
                            ]
                        },
                    },
                ],
            }
        }

        user_groups = list(UserGroup(self.client).get_user_groups())
        execute = self.client.execute.call_args[0]

        assert "GetUserGroupsPyApi" in execute[0]
        assert len(user_groups) == 3

        # Check the attributes of the first user group
        assert user_groups[0].id == "group_id_1"
        assert user_groups[0].name == "Group 1"
        assert user_groups[0].color == UserGroupColor.BLUE
        assert len(user_groups[0].projects) == 2
        assert len(user_groups[0].users) == 2

        # Check the attributes of the second user group
        assert user_groups[1].id == "group_id_2"
        assert user_groups[1].name == "Group 2"
        assert user_groups[1].color == UserGroupColor.BLUE
        assert len(user_groups[1].projects) == 2
        assert len(user_groups[1].users) == 2

        # Check the attributes of the third user group
        assert user_groups[2].id == "group_id_3"
        assert user_groups[2].name == "Group 3"
        assert user_groups[2].color == UserGroupColor.BLUE
        assert len(user_groups[2].projects) == 2
        assert len(user_groups[2].users) == 2


if __name__ == "__main__":
    import subprocess

    subprocess.call(["pytest", "-v", __file__])



import json
from datetime import datetime
from typing import Any

from lbox.exceptions import LabelboxError, ResourceNotFoundError

from labelbox.schema.labeling_service_dashboard import LabelingServiceDashboard
from labelbox.schema.labeling_service_status import LabelingServiceStatus
from labelbox.utils import _CamelCaseMixin

from ..annotated_types import Cuid


class LabelingService(_CamelCaseMixin):
    """
    Labeling service for a project. This is a service that can be requested to label data for a project.
    """

    id: Cuid
    project_id: Cuid
    created_at: datetime
    updated_at: datetime
    created_by_id: Cuid
    status: LabelingServiceStatus
    client: Any  # type Any to avoid circular import from client

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def start(cls, client, project_id: Cuid) -> "LabelingService":
        """
        Starts the labeling service for the project. This is equivalent to a UI action to Request Specialized Labelers

        Returns:
            LabelingService: The labeling service for the project.
        Raises:
            Exception: If the service fails to start.
        """
        query_str = """mutation CreateProjectBoostWorkforcePyApi($projectId: ID!) {
          upsertProjectBoostWorkforce(data: { projectId: $projectId }) {
            success
          }
        }"""
        result = client.execute(query_str, {"projectId": project_id})
        success = result["upsertProjectBoostWorkforce"]["success"]
        if not success:
            raise Exception("Failed to start labeling service")
        return cls.get(client, project_id)

    @classmethod
    def get(cls, client, project_id: Cuid) -> "LabelingService":
        """
        Returns the labeling service associated with the project.

        Raises:
            ResourceNotFoundError: If the project does not have a labeling service.
        """
        query = """
            query GetProjectBoostWorkforcePyApi($projectId: ID!) {
            projectBoostWorkforce(data: { projectId: $projectId }) {
                    id
                    projectId
                    createdAt
                    updatedAt
                    createdById
                    status
                }
            }
        """
        result = client.execute(query, {"projectId": project_id})
        if result["projectBoostWorkforce"] is None:
            raise ResourceNotFoundError(
                message="The project does not have a labeling service."
            )
        data = result["projectBoostWorkforce"]
        data["client"] = client
        return LabelingService(**data)

    def request(self) -> "LabelingService":
        """
        Creates a request to labeling service to start labeling for the project.
        Our back end will validate that the project is ready for labeling and then request the labeling service.

        Returns:
            LabelingService: The labeling service for the project.
        Raises:
            ResourceNotFoundError: If ontology is not associated with the project
                or if any projects required prerequisites are missing.

        """

        query_str = """mutation ValidateAndRequestProjectBoostWorkforcePyApi($projectId: ID!) {
            validateAndRequestProjectBoostWorkforce(
                data: { projectId: $projectId }
            ) {
                success
            }
        }
        """
        result = self.client.execute(
            query_str,
            {"projectId": self.project_id},
            raise_return_resource_not_found=True,
            error_handlers={"MALFORMED_REQUEST": self._raise_readable_errors},
        )
        success = result["validateAndRequestProjectBoostWorkforce"]["success"]
        if not success:
            raise Exception("Failed to start labeling service")
        return LabelingService.get(self.client, self.project_id)

    def _raise_readable_errors(self, response):
        errors = response.json().get("errors", [])
        if errors:
            message = errors[0].get(
                "errors", json.dumps([{"error": "Unknown error"}])
            )
            error_messages = [error["error"] for error in message]
        else:
            error_messages = ["Uknown error"]
        raise LabelboxError(". ".join(error_messages))

    @classmethod
    def getOrCreate(cls, client, project_id: Cuid) -> "LabelingService":
        """
        Returns the labeling service associated with the project. If the project does not have a labeling service, it will create one.

        Returns:
            LabelingService: The labeling service for the project.
        """
        try:
            return cls.get(client, project_id)
        except ResourceNotFoundError:
            return cls.start(client, project_id)

    def dashboard(self) -> LabelingServiceDashboard:
        """
        Returns the dashboard for the labeling service associated with the project.

        Raises:
            ResourceNotFoundError: If the project does not have a labeling service.
        """
        return LabelingServiceDashboard.get(self.client, self.project_id)



import time


from labelbox import DataRow, ExportTask, StreamType


class TestExportDataRow:
    def test_with_data_row_object(
        self, client, data_row, wait_for_data_row_processing
    ):
        data_row = wait_for_data_row_processing(client, data_row)
        time.sleep(7)  # temp fix for ES indexing delay
        export_task = DataRow.export(
            client=client,
            data_rows=[data_row],
            task_name="TestExportDataRow:test_with_data_row_object",
        )
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert isinstance(export_task, ExportTask)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) == 1
        assert (
            list(export_task.get_buffered_stream())[0].json["data_row"]["id"]
            == data_row.uid
        )

    def test_with_data_row_object_buffered(
        self, client, data_row, wait_for_data_row_processing
    ):
        data_row = wait_for_data_row_processing(client, data_row)
        time.sleep(7)  # temp fix for ES indexing delay
        export_task = DataRow.export(
            client=client,
            data_rows=[data_row],
            task_name="TestExportDataRow:test_with_data_row_object_buffered",
        )
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert isinstance(export_task, ExportTask)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) == 1
        assert (
            list(export_task.get_buffered_stream())[0].json["data_row"]["id"]
            == data_row.uid
        )

    def test_with_id(self, client, data_row, wait_for_data_row_processing):
        data_row = wait_for_data_row_processing(client, data_row)
        time.sleep(7)  # temp fix for ES indexing delay
        export_task = DataRow.export(
            client=client,
            data_rows=[data_row.uid],
            task_name="TestExportDataRow:test_with_id",
        )
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert isinstance(export_task, ExportTask)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) == 1
        assert (
            list(export_task.get_buffered_stream())[0].json["data_row"]["id"]
            == data_row.uid
        )

    def test_with_global_key(
        self, client, data_row, wait_for_data_row_processing
    ):
        data_row = wait_for_data_row_processing(client, data_row)
        time.sleep(7)  # temp fix for ES indexing delay
        export_task = DataRow.export(
            client=client,
            global_keys=[data_row.global_key],
            task_name="TestExportDataRow:test_with_global_key",
        )
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert isinstance(export_task, ExportTask)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) == 1
        assert (
            list(export_task.get_buffered_stream())[0].json["data_row"]["id"]
            == data_row.uid
        )

    def test_with_invalid_id(self, client):
        export_task = DataRow.export(
            client=client,
            data_rows=["invalid_id"],
            task_name="TestExportDataRow:test_with_invalid_id",
        )
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert isinstance(export_task, ExportTask)
        assert export_task.has_result() is False
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT)
            is None
        )
        assert (
            export_task.get_total_lines(stream_type=StreamType.RESULT) is None
        )



import uuid
import pytest
from unittest.mock import MagicMock, patch

from labelbox.schema.annotation_import import LabelImport, logger


def test_should_warn_user_about_unsupported_confidence():
    """this test should check running state only to validate running, not completed"""
    id = str(uuid.uuid4())

    labels = [
        {
            "uuid": "b862c586-8614-483c-b5e6-82810f70cac0",
            "schemaId": "ckrazcueb16og0z6609jj7y3y",
            "dataRow": {"id": "ckrazctum0z8a0ybc0b0o0g0v"},
            "confidence": 0.851,
            "bbox": {"top": 1352, "left": 2275, "height": 350, "width": 139},
        },
    ]
    with patch.object(LabelImport, "_create_label_import_from_bytes"):
        with patch.object(logger, "warning") as warning_mock:
            LabelImport.create_from_objects(
                client=MagicMock(), project_id=id, name=id, labels=labels
            )
            warning_mock.assert_called_once()
            "Confidence scores are not supported in Label Import" in warning_mock.call_args_list[
                0
            ].args[0]


def test_invalid_labels_format():
    """this test should confirm that labels are required to be in a form of list"""
    id = str(uuid.uuid4())

    label = {
        "uuid": "b862c586-8614-483c-b5e6-82810f70cac0",
        "schemaId": "ckrazcueb16og0z6609jj7y3y",
        "dataRow": {"id": "ckrazctum0z8a0ybc0b0o0g0v"},
        "bbox": {"top": 1352, "left": 2275, "height": 350, "width": 139},
    }
    with patch.object(LabelImport, "_create_label_import_from_bytes"):
        with pytest.raises(TypeError):
            LabelImport.create_from_objects(
                client=MagicMock(), project_id=id, name=id, labels=label
            )



from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import (
    Label,
    ClassificationAnnotation,
    Radio,
    ClassificationAnswer,
)


def test_generic_data_row_global_key_included():
    expected = [
        {
            "answer": {"schemaId": "ckrb1sfl8099g0y91cxbd5ftb"},
            "dataRow": {"globalKey": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
            "schemaId": "ckrb1sfjx099a0y914hl319ie",
            "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
        }
    ]

    label = Label(
        data=GenericDataRowData(
            global_key="ckrb1sf1i1g7i0ybcdc6oc8ct",
        ),
        annotations=[
            ClassificationAnnotation(
                feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                value=Radio(
                    answer=ClassificationAnswer(
                        feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                    ),
                ),
            )
        ],
    )

    res = list(NDJsonConverter.serialize([label]))

    assert res == expected


def test_dict_data_row_global_key_included():
    expected = [
        {
            "answer": {"schemaId": "ckrb1sfl8099g0y91cxbd5ftb"},
            "dataRow": {"globalKey": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
            "schemaId": "ckrb1sfjx099a0y914hl319ie",
            "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
        }
    ]

    label = Label(
        data={
            "global_key": "ckrb1sf1i1g7i0ybcdc6oc8ct",
        },
        annotations=[
            ClassificationAnnotation(
                feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                value=Radio(
                    answer=ClassificationAnswer(
                        feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                    ),
                ),
            )
        ],
    )

    res = list(NDJsonConverter.serialize([label]))

    assert res == expected



import json

from unittest.mock import MagicMock

from labelbox.schema.internal.descriptor_file_creator import (
    DescriptorFileCreator,
)


def test_chunk_down_by_bytes_row_too_large():
    client = MagicMock()

    descriptor_file_creator = DescriptorFileCreator(client)

    chunk = [{"row_data": "a"}]
    max_chunk_size_bytes = 1

    res = descriptor_file_creator._chunk_down_by_bytes(
        chunk, max_chunk_size_bytes
    )
    assert [x for x in res] == [json.dumps([{"row_data": "a"}])]


def test_chunk_down_by_bytes_more_chunks():
    client = MagicMock()

    descriptor_file_creator = DescriptorFileCreator(client)

    chunk = [{"row_data": "a"}, {"row_data": "b"}]
    max_chunk_size_bytes = len(json.dumps(chunk).encode("utf-8")) - 1

    res = descriptor_file_creator._chunk_down_by_bytes(
        chunk, max_chunk_size_bytes
    )
    assert [x for x in res] == [
        json.dumps([{"row_data": "a"}]),
        json.dumps([{"row_data": "b"}]),
    ]


def test_chunk_down_by_bytes_one_chunk():
    client = MagicMock()

    descriptor_file_creator = DescriptorFileCreator(client)

    chunk = [{"row_data": "a"}, {"row_data": "b"}]
    max_chunk_size_bytes = len(json.dumps(chunk).encode("utf-8"))

    res = descriptor_file_creator._chunk_down_by_bytes(
        chunk, max_chunk_size_bytes
    )
    assert [x for x in res] == [
        json.dumps([{"row_data": "a"}, {"row_data": "b"}])
    ]



import pytest
from unittest.mock import MagicMock

from labelbox.schema.project import Project
from labelbox.schema.ontology_kind import EditorTaskType


@pytest.fixture
def project_entity():
    return Project(
        MagicMock(),
        {
            "id": "test",
            "name": "test",
            "createdAt": "2021-06-01T00:00:00.000Z",
            "updatedAt": "2021-06-01T00:00:00.000Z",
            "autoAuditNumberOfLabels": 1,
            "autoAuditPercentage": 100,
            "dataRowCount": 1,
            "description": "test",
            "editorTaskType": "MODEL_CHAT_EVALUATION",
            "lastActivityTime": "2021-06-01T00:00:00.000Z",
            "allowedMediaType": "IMAGE",
            "setupComplete": "2021-06-01T00:00:00.000Z",
            "modelSetupComplete": None,
            "uploadType": "Auto",
            "isBenchmarkEnabled": False,
            "isConsensusEnabled": False,
        },
    )


@pytest.mark.parametrize(
    "api_editor_task_type, expected_editor_task_type",
    [
        (None, EditorTaskType.Missing),
        ("MODEL_CHAT_EVALUATION", EditorTaskType.ModelChatEvaluation),
        ("RESPONSE_CREATION", EditorTaskType.ResponseCreation),
        (
            "OFFLINE_MODEL_CHAT_EVALUATION",
            EditorTaskType.OfflineModelChatEvaluation,
        ),
        ("NEW_TYPE", EditorTaskType.Missing),
    ],
)
def test_project_editor_task_type(
    api_editor_task_type, expected_editor_task_type, project_entity
):
    client = MagicMock()
    project = Project(
        client,
        {
            "id": "test",
            "name": "test",
            "createdAt": "2021-06-01T00:00:00.000Z",
            "updatedAt": "2021-06-01T00:00:00.000Z",
            "autoAuditNumberOfLabels": 1,
            "autoAuditPercentage": 100,
            "dataRowCount": 1,
            "description": "test",
            "editorTaskType": api_editor_task_type,
            "lastActivityTime": "2021-06-01T00:00:00.000Z",
            "allowedMediaType": "IMAGE",
            "setupComplete": "2021-06-01T00:00:00.000Z",
            "modelSetupComplete": None,
            "uploadType": "Auto",
            "isBenchmarkEnabled": False,
            "isConsensusEnabled": False,
        },
    )

    assert project.editor_task_type == expected_editor_task_type



import pytest
import cv2

from labelbox.data.annotation_types import Polygon, Point
from pydantic import ValidationError


def test_polygon():
    with pytest.raises(ValidationError):
        polygon = Polygon()

    with pytest.raises(ValidationError):
        polygon = Polygon(points=[[0, 1], [2, 3]])

    with pytest.raises(ValidationError):
        polygon = Polygon(points=[Point(x=0, y=1), Point(x=0, y=1)])

    points = [[0.0, 1.0], [0.0, 2.0], [2.0, 2.0], [2.0, 0.0]]
    expected = {"coordinates": [points + [points[0]]], "type": "Polygon"}
    polygon = Polygon(points=[Point(x=x, y=y) for x, y in points])
    assert polygon.geometry == expected
    expected["coordinates"] = tuple(
        [tuple([tuple(x) for x in points + [points[0]]])]
    )
    assert polygon.shapely.__geo_interface__ == expected

    raster = polygon.draw(10, 10)
    assert (cv2.imread("tests/data/assets/polygon.png") == raster).all()



import datetime
from enum import Enum
from typing import Annotated, List, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PlainSerializer,
    field_validator,
)

from labelbox.schema.labeling_service_status import LabelingServiceStatus
from labelbox.utils import format_iso_datetime


class BaseSearchFilter(BaseModel):
    """
    Shared code for all search filters
    """

    model_config = ConfigDict(use_enum_values=True)


class OperationTypeEnum(Enum):
    """
    Supported search entity types
    Each type corresponds to a different filter class
    """

    Organization = "organization_id"
    SharedWithOrganization = "shared_with_organizations"
    Workspace = "workspace"
    Tag = "tag"
    Stage = "stage"
    WorforceRequestedDate = "workforce_requested_at"
    WorkforceStageUpdatedDate = "workforce_stage_updated_at"
    TaskCompletedCount = "task_completed_count"
    TaskRemainingCount = "task_remaining_count"


def convert_enum_to_str(enum_or_str: Union[Enum, str]) -> str:
    if isinstance(enum_or_str, Enum):
        return enum_or_str.value
    return enum_or_str


OperationType = Annotated[
    OperationTypeEnum, PlainSerializer(convert_enum_to_str, return_type=str)
]

IsoDatetimeType = Annotated[
    datetime.datetime, PlainSerializer(format_iso_datetime)
]


class IdOperator(Enum):
    """
    Supported operators for ids like org ids, workspace ids, etc
    """

    Is = "is"


class RangeOperatorWithSingleValue(Enum):
    """
    Supported operators for dates
    """

    Equals = "EQUALS"
    GreaterThanOrEqual = "GREATER_THAN_OR_EQUAL"
    LessThanOrEqual = "LESS_THAN_OR_EQUAL"


class RangeDateTimeOperatorWithSingleValue(Enum):
    """
    Supported operators for dates
    """

    GreaterThanOrEqual = "GREATER_THAN_OR_EQUAL"
    LessThanOrEqual = "LESS_THAN_OR_EQUAL"


class RangeOperatorWithValue(Enum):
    """
    Supported operators for date ranges
    """

    Between = "BETWEEN"


class OrganizationFilter(BaseSearchFilter):
    """
    Filter for organization to which projects belong
    """

    operation: OperationType = Field(
        default=OperationType.Organization, serialization_alias="type"
    )
    operator: IdOperator
    values: List[str]


class SharedWithOrganizationFilter(BaseSearchFilter):
    """
    Find project shared with the organization (i.e. not having this organization as a tenantId)
    """

    operation: OperationType = Field(
        default=OperationType.SharedWithOrganization, serialization_alias="type"
    )
    operator: IdOperator
    values: List[str]


class WorkspaceFilter(BaseSearchFilter):
    """
    Filter for workspace
    """

    operation: OperationType = Field(
        default=OperationType.Workspace, serialization_alias="type"
    )
    operator: IdOperator
    values: List[str]


class TagFilter(BaseSearchFilter):
    """
    Filter for project tags
    values are tag ids
    """

    operation: OperationType = Field(
        default=OperationType.Tag, serialization_alias="type"
    )
    operator: IdOperator
    values: List[str]


class ProjectStageFilter(BaseSearchFilter):
    """
    Filter labelbox service / aka project stages
        Stages are: requested, in_progress, completed etc. as described by LabelingServiceStatus
    """

    operation: OperationType = Field(
        default=OperationType.Stage, serialization_alias="type"
    )
    operator: IdOperator
    values: List[LabelingServiceStatus]

    @field_validator("values", mode="before")
    def validate_values(cls, values):
        disallowed_values = [LabelingServiceStatus.Missing]
        for value in values:
            if value in disallowed_values:
                raise ValueError(
                    f"{value} is not a valid value for ProjectStageFilter"
                )

        return values


class DateValue(BaseSearchFilter):
    """
    Date value for a search filter

    Date formats:
        datetime: an existing datetime object
        str the following formats are accepted: YYYY-MM-DD[T]HH:MM[:SS[.ffffff]][Z or [±]HH[:]MM]
        NOTE
            if a date / datetime string is passed without a timezone, we will assume the time is UTC and convert it to a local timezone
            so for a string '2024-01-01' that is run on a computer in PST, we would convert it to '2024-01-01T08:00:00Z'
            while the same string in EST will get converted to '2024-01-01T05:00:00Z'
    """

    operator: RangeDateTimeOperatorWithSingleValue
    value: IsoDatetimeType


class IntegerValue(BaseSearchFilter):
    operator: RangeOperatorWithSingleValue
    value: int


class WorkforceStageUpdatedFilter(BaseSearchFilter):
    """
    Filter for workforce stage updated date
    """

    operation: OperationType = Field(
        default=OperationType.WorkforceStageUpdatedDate,
        serialization_alias="type",
    )
    value: DateValue


class WorkforceRequestedDateFilter(BaseSearchFilter):
    """
    Filter for workforce requested date
    """

    operation: OperationType = Field(
        default=OperationType.WorforceRequestedDate, serialization_alias="type"
    )
    value: DateValue


class DateRange(BaseSearchFilter):
    """
    Date range for a search filter
    """

    min: IsoDatetimeType
    max: IsoDatetimeType


class DateRangeValue(BaseSearchFilter):
    """
    Date range value for a search filter
    """

    operator: RangeOperatorWithValue
    value: DateRange


class WorkforceRequestedDateRangeFilter(BaseSearchFilter):
    """
    Filter for workforce requested date range
    """

    operation: OperationType = Field(
        default=OperationType.WorforceRequestedDate, serialization_alias="type"
    )
    value: DateRangeValue


class WorkforceStageUpdatedRangeFilter(BaseSearchFilter):
    """
    Filter for workforce stage updated date range
    """

    operation: OperationType = Field(
        default=OperationType.WorkforceStageUpdatedDate,
        serialization_alias="type",
    )
    value: DateRangeValue


class TaskCompletedCountFilter(BaseSearchFilter):
    """
    Filter for completed tasks count
        A task maps to a data row. Task completed should map to a data row in a labeling queue DONE
    """

    operation: OperationType = Field(
        default=OperationType.TaskCompletedCount, serialization_alias="type"
    )
    value: IntegerValue


class TaskRemainingCountFilter(BaseSearchFilter):
    """
    Filter for remaining tasks count. Reverse of TaskCompletedCountFilter
    """

    operation: OperationType = Field(
        default=OperationType.TaskRemainingCount, serialization_alias="type"
    )
    value: IntegerValue


SearchFilter = Union[
    OrganizationFilter,
    WorkspaceFilter,
    SharedWithOrganizationFilter,
    TagFilter,
    ProjectStageFilter,
    WorkforceRequestedDateFilter,
    WorkforceStageUpdatedFilter,
    WorkforceRequestedDateRangeFilter,
    WorkforceStageUpdatedRangeFilter,
    TaskCompletedCountFilter,
    TaskRemainingCountFilter,
]


def _dict_to_graphql_string(d: Union[dict, list, str, int]) -> str:
    if isinstance(d, dict):
        return (
            "{"
            + ", ".join(
                f"{k}: {_dict_to_graphql_string(v)}" for k, v in d.items()
            )
            + "}"
        )
    elif isinstance(d, list):
        return (
            "[" + ", ".join(_dict_to_graphql_string(item) for item in d) + "]"
        )
    else:
        return f'"{d}"' if isinstance(d, str) else str(d)


def build_search_filter(filter: List[SearchFilter]):
    """
    Converts a list of search filters to a graphql string
    """
    filters = [
        _dict_to_graphql_string(f.model_dump(by_alias=True)) for f in filter
    ]
    return "[" + ", ".join(filters) + "]"



def remove_keys_recursive(d, keys):
    for k in keys:
        if k in d:
            del d[k]
    for k, v in d.items():
        if isinstance(v, dict):
            remove_keys_recursive(v, keys)
        elif isinstance(v, list):
            for i in v:
                if isinstance(i, dict):
                    remove_keys_recursive(i, keys)


# NOTE this uses quite a primitive check for cuids but I do not think it is worth coming up with a better one
# Also this function is NOT written with performance in mind, good for small to mid size dicts like we have in our test
def rename_cuid_key_recursive(d):
    new_key = "<cuid>"
    for k in list(d.keys()):
        if len(k) == 25 and not k.isalpha():  # primitive check for cuid
            d[new_key] = d.pop(k)
    for k, v in d.items():
        if isinstance(v, dict):
            rename_cuid_key_recursive(v)
        elif isinstance(v, list):
            for i in v:
                if isinstance(i, dict):
                    rename_cuid_key_recursive(i)


INTEGRATION_SNAPSHOT_DIRECTORY = "tests/integration/snapshots"



from copy import copy

import pytest

from labelbox.schema.dataset import Dataset


@pytest.fixture
def data_for_dataset_order_test(client, rand_gen):
    name = rand_gen(str)
    dataset1 = client.create_dataset(name=name)
    dataset2 = client.create_dataset(name=name)

    yield name

    dataset1.delete()
    dataset2.delete()


def test_get_one_and_many_dataset_order(client, data_for_dataset_order_test):
    name = data_for_dataset_order_test

    paginator = client.get_datasets(where=Dataset.name == name)
    # confirm get_one returns first dataset
    all_datasets = list(paginator)
    assert len(all_datasets) == 2
    get_one_dataset = copy(paginator).get_one()
    assert get_one_dataset.uid == all_datasets[0].uid

    # confirm get_many(1) returns first dataset
    get_many_datasets = copy(paginator).get_many(1)
    assert get_many_datasets[0].uid == all_datasets[0].uid



import base64
from io import BytesIO
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from PIL import Image
from pydantic import BaseModel

from labelbox.data.annotation_types.data import GenericDataRowData
from labelbox.data.annotation_types.data.raster import MaskData
from labelbox.data.annotation_types.ner.conversation_entity import (
    ConversationEntity,
)
from labelbox.data.annotation_types.video import (
    VideoObjectAnnotation,
)
from labelbox.data.mixins import (
    ConfidenceMixin,
    CustomMetric,
    CustomMetricsMixin,
    CustomMetricsNotSupportedMixin,
)

from ....annotated_types import Cuid
from ...annotation_types.annotation import (
    ClassificationAnnotation,
    ObjectAnnotation,
)
from ...annotation_types.data import GenericDataRowData
from ...annotation_types.geometry import (
    DocumentRectangle,
    Line,
    Mask,
    Point,
    Polygon,
    Rectangle,
)
from ...annotation_types.ner import (
    DocumentEntity,
    DocumentTextSelection,
    TextEntity,
)
from ...annotation_types.video import (
    MaskFrame,
    MaskInstance,
    VideoMaskAnnotation,
)
from .base import DataRow, NDAnnotation, NDJsonBase
from .classification import (
    NDSubclassification,
    NDSubclassificationType,
)


class NDBaseObject(NDAnnotation):
    classifications: List[NDSubclassificationType] = []


class VideoSupported(BaseModel):
    # support for video for objects are per-frame basis
    frame: int


class _Point(BaseModel):
    x: float
    y: float


class Bbox(BaseModel):
    top: float
    left: float
    height: float
    width: float


class NDPoint(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    point: _Point

    def to_common(self) -> Point:
        return Point(x=self.point.x, y=self.point.y)

    @classmethod
    def from_common(
        cls,
        uuid: str,
        point: Point,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDPoint":
        return cls(
            point={"x": point.x, "y": point.y},
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDFramePoint(VideoSupported):
    point: _Point
    classifications: List[NDSubclassificationType] = []

    def to_common(
        self, name: str, feature_schema_id: Cuid, segment_index: int
    ) -> VideoObjectAnnotation:
        return VideoObjectAnnotation(
            frame=self.frame,
            segment_index=segment_index,
            keyframe=True,
            name=name,
            feature_schema_id=feature_schema_id,
            value=Point(x=self.point.x, y=self.point.y),
            classifications=[
                NDSubclassification.to_common(annot)
                for annot in self.classifications
            ],
        )

    @classmethod
    def from_common(
        cls,
        frame: int,
        point: Point,
        classifications: List[NDSubclassificationType],
    ):
        return cls(
            frame=frame,
            point=_Point(x=point.x, y=point.y),
            classifications=classifications,
        )


class NDLine(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    line: List[_Point]

    def to_common(self) -> Line:
        return Line(points=[Point(x=pt.x, y=pt.y) for pt in self.line])

    @classmethod
    def from_common(
        cls,
        uuid: str,
        line: Line,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDLine":
        return cls(
            line=[{"x": pt.x, "y": pt.y} for pt in line.points],
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDFrameLine(VideoSupported):
    line: List[_Point]
    classifications: List[NDSubclassificationType] = []

    def to_common(
        self, name: str, feature_schema_id: Cuid, segment_index: int
    ) -> VideoObjectAnnotation:
        return VideoObjectAnnotation(
            frame=self.frame,
            segment_index=segment_index,
            keyframe=True,
            name=name,
            feature_schema_id=feature_schema_id,
            value=Line(points=[Point(x=pt.x, y=pt.y) for pt in self.line]),
            classifications=[
                NDSubclassification.to_common(annot)
                for annot in self.classifications
            ],
        )

    @classmethod
    def from_common(
        cls,
        frame: int,
        line: Line,
        classifications: List[NDSubclassificationType],
    ):
        return cls(
            frame=frame,
            line=[{"x": pt.x, "y": pt.y} for pt in line.points],
            classifications=classifications,
        )


class NDPolygon(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    polygon: List[_Point]

    def to_common(self) -> Polygon:
        return Polygon(points=[Point(x=pt.x, y=pt.y) for pt in self.polygon])

    @classmethod
    def from_common(
        cls,
        uuid: str,
        polygon: Polygon,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDPolygon":
        return cls(
            polygon=[{"x": pt.x, "y": pt.y} for pt in polygon.points],
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDRectangle(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    bbox: Bbox

    def to_common(self) -> Rectangle:
        return Rectangle(
            start=Point(x=self.bbox.left, y=self.bbox.top),
            end=Point(
                x=self.bbox.left + self.bbox.width,
                y=self.bbox.top + self.bbox.height,
            ),
        )

    @classmethod
    def from_common(
        cls,
        uuid: str,
        rectangle: Rectangle,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDRectangle":
        return cls(
            bbox=Bbox(
                top=min(rectangle.start.y, rectangle.end.y),
                left=min(rectangle.start.x, rectangle.end.x),
                height=abs(rectangle.end.y - rectangle.start.y),
                width=abs(rectangle.end.x - rectangle.start.x),
            ),
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            page=extra.get("page"),
            unit=extra.get("unit"),
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDDocumentRectangle(NDRectangle):
    page: int
    unit: str

    def to_common(self) -> DocumentRectangle:
        return DocumentRectangle(
            start=Point(x=self.bbox.left, y=self.bbox.top),
            end=Point(
                x=self.bbox.left + self.bbox.width,
                y=self.bbox.top + self.bbox.height,
            ),
            page=self.page,
            unit=self.unit,
        )

    @classmethod
    def from_common(
        cls,
        uuid: str,
        rectangle: DocumentRectangle,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDRectangle":
        return cls(
            bbox=Bbox(
                top=min(rectangle.start.y, rectangle.end.y),
                left=min(rectangle.start.x, rectangle.end.x),
                height=abs(rectangle.end.y - rectangle.start.y),
                width=abs(rectangle.end.x - rectangle.start.x),
            ),
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            page=rectangle.page,
            unit=rectangle.unit.value,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDFrameRectangle(VideoSupported):
    bbox: Bbox
    classifications: List[NDSubclassificationType] = []

    def to_common(
        self, name: str, feature_schema_id: Cuid, segment_index: int
    ) -> VideoObjectAnnotation:
        return VideoObjectAnnotation(
            frame=self.frame,
            segment_index=segment_index,
            keyframe=True,
            name=name,
            feature_schema_id=feature_schema_id,
            value=Rectangle(
                start=Point(x=self.bbox.left, y=self.bbox.top),
                end=Point(
                    x=self.bbox.left + self.bbox.width,
                    y=self.bbox.top + self.bbox.height,
                ),
            ),
            classifications=[
                NDSubclassification.to_common(annot)
                for annot in self.classifications
            ],
        )

    @classmethod
    def from_common(
        cls,
        frame: int,
        rectangle: Rectangle,
        classifications: List[NDSubclassificationType],
    ):
        return cls(
            frame=frame,
            bbox=Bbox(
                top=min(rectangle.start.y, rectangle.end.y),
                left=min(rectangle.start.x, rectangle.end.x),
                height=abs(rectangle.end.y - rectangle.start.y),
                width=abs(rectangle.end.x - rectangle.start.x),
            ),
            classifications=classifications,
        )


class NDSegment(BaseModel):
    keyframes: List[Union[NDFrameRectangle, NDFramePoint, NDFrameLine]]

    @staticmethod
    def lookup_segment_object_type(segment: List) -> "NDFrameObjectType":
        """Used for determining which object type the annotation contains
        returns the object type"""
        result = {
            Rectangle: NDFrameRectangle,
            Point: NDFramePoint,
            Line: NDFrameLine,
        }.get(type(segment[0].value))
        return result

    @staticmethod
    def segment_with_uuid(
        keyframe: Union[NDFrameRectangle, NDFramePoint, NDFrameLine], uuid: str
    ):
        keyframe._uuid = uuid
        keyframe.extra = {"uuid": uuid}
        return keyframe

    def to_common(
        self, name: str, feature_schema_id: Cuid, uuid: str, segment_index: int
    ):
        return [
            self.segment_with_uuid(
                keyframe.to_common(
                    name=name,
                    feature_schema_id=feature_schema_id,
                    segment_index=segment_index,
                ),
                uuid,
            )
            for keyframe in self.keyframes
        ]

    @classmethod
    def from_common(cls, segment):
        nd_frame_object_type = cls.lookup_segment_object_type(segment)

        return cls(
            keyframes=[
                nd_frame_object_type.from_common(
                    object_annotation.frame,
                    object_annotation.value,
                    [
                        NDSubclassification.from_common(annot)
                        for annot in object_annotation.classifications
                    ],
                )
                for object_annotation in segment
            ]
        )


class NDSegments(NDBaseObject):
    segments: List[NDSegment]

    def to_common(self, name: str, feature_schema_id: Cuid):
        result = []
        for idx, segment in enumerate(self.segments):
            result.extend(
                segment.to_common(
                    name=name,
                    feature_schema_id=feature_schema_id,
                    segment_index=idx,
                    uuid=self.uuid,
                )
            )
        return result

    @classmethod
    def from_common(
        cls,
        segments: List[VideoObjectAnnotation],
        data: GenericDataRowData,
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
    ) -> "NDSegments":
        segments = [NDSegment.from_common(segment) for segment in segments]

        return cls(
            segments=segments,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=extra.get("uuid"),
        )


class _URIMask(BaseModel):
    instanceURI: str
    colorRGB: Tuple[int, int, int]


class _PNGMask(BaseModel):
    png: str


class NDMask(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    mask: Union[_URIMask, _PNGMask]

    def to_common(self) -> Mask:
        if isinstance(self.mask, _URIMask):
            return Mask(
                mask=MaskData(url=self.mask.instanceURI),
                color=self.mask.colorRGB,
            )
        else:
            encoded_image_bytes = self.mask.png.encode("utf-8")
            image_bytes = base64.b64decode(encoded_image_bytes)
            image = np.array(Image.open(BytesIO(image_bytes)))
            if np.max(image) > 1:
                raise ValueError(
                    f"Expected binary mask. Found max value of {np.max(image)}"
                )
            # Color is 1,1,1 because it is a binary array and we are just stacking it into 3 channels
            return Mask(mask=MaskData.from_2D_arr(image), color=(1, 1, 1))

    @classmethod
    def from_common(
        cls,
        uuid: str,
        mask: Mask,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDMask":
        if mask.mask.url is not None:
            lbv1_mask = _URIMask(instanceURI=mask.mask.url, colorRGB=mask.color)
        else:
            binary = np.all(mask.mask.value == mask.color, axis=-1)
            im_bytes = BytesIO()
            Image.fromarray(binary, "L").save(im_bytes, format="PNG")
            lbv1_mask = _PNGMask(
                png=base64.b64encode(im_bytes.getvalue()).decode("utf-8")
            )

        return cls(
            mask=lbv1_mask,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDVideoMasksFramesInstances(BaseModel):
    frames: List[MaskFrame]
    instances: List[MaskInstance]


class NDVideoMasks(
    NDJsonBase,
    ConfidenceMixin,
    CustomMetricsNotSupportedMixin,
):
    masks: NDVideoMasksFramesInstances

    def to_common(self) -> VideoMaskAnnotation:
        for mask_frame in self.masks.frames:
            if mask_frame.im_bytes:
                mask_frame.im_bytes = base64.b64decode(
                    mask_frame.im_bytes.encode("utf-8")
                )

        return VideoMaskAnnotation(
            frames=self.masks.frames,
            instances=self.masks.instances,
        )

    @classmethod
    def from_common(cls, annotation, data):
        for mask_frame in annotation.frames:
            if mask_frame.im_bytes:
                mask_frame.im_bytes = base64.b64encode(
                    mask_frame.im_bytes
                ).decode("utf-8")

        return cls(
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            masks=NDVideoMasksFramesInstances(
                frames=annotation.frames, instances=annotation.instances
            ),
        )


class Location(BaseModel):
    start: int
    end: int


class NDTextEntity(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    location: Location

    def to_common(self) -> TextEntity:
        return TextEntity(start=self.location.start, end=self.location.end)

    @classmethod
    def from_common(
        cls,
        uuid: str,
        text_entity: TextEntity,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDTextEntity":
        return cls(
            location=Location(
                start=text_entity.start,
                end=text_entity.end,
            ),
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDDocumentEntity(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    name: str
    text_selections: List[DocumentTextSelection]

    def to_common(self) -> DocumentEntity:
        return DocumentEntity(
            name=self.name, text_selections=self.text_selections
        )

    @classmethod
    def from_common(
        cls,
        uuid: str,
        document_entity: DocumentEntity,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDDocumentEntity":
        return cls(
            text_selections=document_entity.text_selections,
            dataRow=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDConversationEntity(NDTextEntity):
    message_id: str

    def to_common(self) -> ConversationEntity:
        return ConversationEntity(
            start=self.location.start,
            end=self.location.end,
            message_id=self.message_id,
        )

    @classmethod
    def from_common(
        cls,
        uuid: str,
        conversation_entity: ConversationEntity,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDConversationEntity":
        return cls(
            location=Location(
                start=conversation_entity.start, end=conversation_entity.end
            ),
            message_id=conversation_entity.message_id,
            dataRow=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            classifications=classifications,
            confidence=confidence,
            custom_metrics=custom_metrics,
        )


class NDObject:
    @staticmethod
    def to_common(annotation: "NDObjectType") -> ObjectAnnotation:
        common_annotation = annotation.to_common()
        classifications = [
            NDSubclassification.to_common(annot)
            for annot in annotation.classifications
        ]
        confidence = (
            annotation.confidence if hasattr(annotation, "confidence") else None
        )

        custom_metrics = (
            annotation.custom_metrics
            if hasattr(annotation, "custom_metrics")
            else None
        )
        return ObjectAnnotation(
            value=common_annotation,
            name=annotation.name,
            feature_schema_id=annotation.schema_id,
            classifications=classifications,
            extra={
                "uuid": annotation.uuid,
                "page": annotation.page,
                "unit": annotation.unit,
            },
            confidence=confidence,
            custom_metrics=custom_metrics,
        )

    @classmethod
    def from_common(
        cls,
        annotation: Union[
            ObjectAnnotation,
            List[List[VideoObjectAnnotation]],
            VideoMaskAnnotation,
        ],
        data: GenericDataRowData,
    ) -> Union[
        NDLine,
        NDPoint,
        NDPolygon,
        NDDocumentRectangle,
        NDRectangle,
        NDMask,
        NDTextEntity,
    ]:
        obj = cls.lookup_object(annotation)

        # if it is video segments
        if obj == NDSegments:
            first_video_annotation = annotation[0][0]
            args = dict(
                segments=annotation,
                data=data,
                name=first_video_annotation.name,
                feature_schema_id=first_video_annotation.feature_schema_id,
                extra=first_video_annotation.extra,
            )

            return obj.from_common(**args)
        elif obj == NDVideoMasks:
            return obj.from_common(annotation, data)

        subclasses = [
            NDSubclassification.from_common(annot)
            for annot in annotation.classifications
        ]
        optional_kwargs = {}
        if annotation.confidence:
            optional_kwargs["confidence"] = annotation.confidence

        if annotation.custom_metrics:
            optional_kwargs["custom_metrics"] = annotation.custom_metrics

        return obj.from_common(
            str(annotation._uuid),
            annotation.value,
            subclasses,
            annotation.name,
            annotation.feature_schema_id,
            annotation.extra,
            data,
            **optional_kwargs,
        )

    @staticmethod
    def lookup_object(
        annotation: Union[ObjectAnnotation, List],
    ) -> "NDObjectType":
        if isinstance(annotation, VideoMaskAnnotation):
            result = NDVideoMasks
        elif isinstance(annotation, list):
            result = NDSegments
        else:
            result = {
                Line: NDLine,
                Point: NDPoint,
                Polygon: NDPolygon,
                Rectangle: NDRectangle,
                DocumentRectangle: NDDocumentRectangle,
                Mask: NDMask,
                TextEntity: NDTextEntity,
                DocumentEntity: NDDocumentEntity,
                ConversationEntity: NDConversationEntity,
            }.get(type(annotation.value))
        if result is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation.value)}`"
            )
        return result


NDEntityType = Union[NDConversationEntity, NDTextEntity]
NDObjectType = Union[
    NDLine,
    NDPolygon,
    NDPoint,
    NDDocumentRectangle,
    NDRectangle,
    NDMask,
    NDEntityType,
    NDDocumentEntity,
]

NDFrameObjectType = NDFrameRectangle, NDFramePoint, NDFrameLine



import json


class NdjsonDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def decode(self, s: str, *args, **kwargs):
        lines = ",".join(s.splitlines())
        text = f"[{lines}]"  # NOTE: this is a hack to make json.loads work for ndjson
        return super().decode(text, *args, **kwargs)


def loads(ndjson_string, **kwargs) -> list:
    kwargs.setdefault("cls", NdjsonDecoder)
    return json.loads(ndjson_string, **kwargs)


def dumps(obj, **kwargs):
    lines = map(lambda obj: json.dumps(obj, **kwargs), obj)
    return "\n".join(lines)


def dump(obj, io, **kwargs):
    lines = dumps(obj, **kwargs)
    io.write(lines)


def reader(io_handle, **kwargs):
    for line in io_handle:
        yield json.loads(line, **kwargs)



import pytest

from labelbox import Project
from labelbox.orm.comparison import Comparison, LogicalExpression


def test_comparison_creation():
    comparison = Comparison.Op.EQ(Project.name, "test")
    assert comparison.op == Comparison.Op.EQ
    assert comparison.field == Project.name
    assert comparison.value == "test"


def test_comparison_equality():
    Op = Comparison.Op
    assert Op.EQ(Project.name, "test") == Op.EQ(Project.name, "test")
    assert Op.EQ(Project.name, "test") != Op.EQ(Project.uid, "test")
    assert Op.EQ(Project.name, "test") != Op.EQ(Project.name, "t")
    assert Op.EQ(Project.name, "test") != Op.NE(Project.name, "test")


def test_rich_comparison():
    Op = Comparison.Op
    assert (Project.uid == "uid") == Op.EQ(Project.uid, "uid")
    assert (Project.uid != "uid") == Op.NE(Project.uid, "uid")
    assert (Project.uid < "uid") == Op.LT(Project.uid, "uid")
    assert (Project.uid <= "uid") == Op.LE(Project.uid, "uid")
    assert (Project.uid > "uid") == Op.GT(Project.uid, "uid")
    assert (Project.uid >= "uid") == Op.GE(Project.uid, "uid")

    # inverse operands
    assert ("uid" == Project.uid) == Op.EQ(Project.uid, "uid")
    assert ("uid" != Project.uid) == Op.NE(Project.uid, "uid")
    assert ("uid" < Project.uid) == Op.GT(Project.uid, "uid")
    assert ("uid" <= Project.uid) == Op.GE(Project.uid, "uid")
    assert ("uid" > Project.uid) == Op.LT(Project.uid, "uid")
    assert ("uid" >= Project.uid) == Op.LE(Project.uid, "uid")


def test_logical_expr_creation():
    comparison_1 = Comparison.Op.EQ(Project.name, "name")
    comparison_2 = Comparison.Op.LT(Project.uid, "uid")

    op = LogicalExpression.Op.AND(comparison_1, comparison_2)
    assert op.op == LogicalExpression.Op.AND
    assert op.first == comparison_1
    assert op.second == comparison_2


def test_logical_expr_ops():
    comparison_1 = Comparison.Op.EQ(Project.name, "name")
    comparison_2 = Comparison.Op.LT(Project.uid, "uid")

    log_op_1 = comparison_1 & comparison_2
    assert log_op_1 == LogicalExpression.Op.AND(comparison_1, comparison_2)
    log_op_2 = log_op_1 | comparison_1
    assert log_op_2 == LogicalExpression.Op.OR(log_op_1, comparison_1)
    log_op_3 = ~log_op_2
    assert log_op_3 == LogicalExpression.Op.NOT(log_op_2)

    # Can't create logical expressions with anything except comparisons and
    # other logical expressions.
    with pytest.raises(TypeError):
        logical_op = comparison_1 & 42
    with pytest.raises(TypeError):
        logical_op = comparison_1 | 42



from labelbox.data.annotation_types.ner.text_entity import TextEntity
from labelbox.utils import _CamelCaseMixin


class ConversationEntity(TextEntity, _CamelCaseMixin):
    """Represents a text entity"""

    message_id: str



from typing import List, Union

from labelbox.data.annotation_types.base_annotation import BaseAnnotation
from labelbox.data.annotation_types.geometry.geometry import Geometry

from labelbox.data.mixins import ConfidenceMixin, CustomMetricsMixin

from labelbox.data.annotation_types.classification.classification import (
    ClassificationAnnotation,
)
from .ner import DocumentEntity, TextEntity, ConversationEntity
from typing import Optional


class ObjectAnnotation(BaseAnnotation, ConfidenceMixin, CustomMetricsMixin):
    """Generic localized annotation (non classifications)

    >>> ObjectAnnotation(
    >>>     value=Rectangle(
    >>>        start=Point(x=0, y=0),
    >>>        end=Point(x=1, y=1)
    >>>     ),
    >>>     feature_schema_id="my-feature-schema-id"
    >>> )

    Args:
        name (Optional[str])
        feature_schema_id (Optional[Cuid])
        value (Union[TextEntity, Geometry]): Localization of the annotation
        classifications (Optional[List[ClassificationAnnotation]]): Optional sub classification of the annotation
        extra (Dict[str, Any])
    """

    value: Union[TextEntity, ConversationEntity, DocumentEntity, Geometry]
    classifications: Optional[List[ClassificationAnnotation]] = []



import pytest
from lbox.exceptions import ResourceNotFoundError


def test_add_single_model_config(live_chat_evaluation_project, model_config):
    configured_project = live_chat_evaluation_project
    project_model_config_id = configured_project.add_model_config(
        model_config.uid
    )

    assert set(
        config.uid for config in configured_project.project_model_configs()
    ) == set([project_model_config_id])

    assert configured_project.delete_project_model_config(
        project_model_config_id
    )


def test_add_multiple_model_config(
    client,
    rand_gen,
    live_chat_evaluation_project,
    model_config,
    valid_model_id,
):
    configured_project = live_chat_evaluation_project
    second_model_config = client.create_model_config(
        rand_gen(str), valid_model_id, {"param": "value"}
    )
    first_project_model_config_id = configured_project.add_model_config(
        model_config.uid
    )
    second_project_model_config_id = configured_project.add_model_config(
        second_model_config.uid
    )
    expected_model_configs = set(
        [first_project_model_config_id, second_project_model_config_id]
    )

    assert (
        set(config.uid for config in configured_project.project_model_configs())
        == expected_model_configs
    )

    for project_model_config_id in expected_model_configs:
        assert configured_project.delete_project_model_config(
            project_model_config_id
        )


def test_delete_project_model_config(
    live_chat_evaluation_project, model_config
):
    configured_project = live_chat_evaluation_project
    assert configured_project.delete_project_model_config(
        configured_project.add_model_config(model_config.uid)
    )
    assert not len(configured_project.project_model_configs())


def test_delete_nonexistant_project_model_config(configured_project):
    with pytest.raises(ResourceNotFoundError):
        configured_project.delete_project_model_config(
            "nonexistant_project_model_config"
        )



from labelbox.data.annotation_types.annotation import ClassificationAnnotation
from labelbox.data.annotation_types.classification.classification import (
    Text,
)
from labelbox.data.annotation_types.data import GenericDataRowData
from labelbox.data.annotation_types.label import Label

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_serialization():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=GenericDataRowData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
        ),
        annotations=[
            ClassificationAnnotation(
                name="radio_question_geo",
                confidence=0.5,
                value=Text(answer="first_radio_answer"),
            )
        ],
    )

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)
    assert (
        "confidence" not in res
    )  # because confidence needs to be set on the annotation itself
    assert res["name"] == "radio_question_geo"
    assert res["answer"] == "first_radio_answer"
    assert res["dataRow"]["id"] == "bkj7z2q0b0000jx6x0q2q7q0d"



from typing import Callable, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from pydantic import field_validator
from shapely.geometry import MultiPolygon, Polygon

from ..data import MaskData
from .geometry import Geometry


class Mask(Geometry):
    """Mask used to represent a single class in a larger segmentation mask

    Example of a mutually exclusive class

    >>> arr = MaskData.from_2D_arr([
    >>>    [0, 0, 0],
    >>>    [1, 1, 1],
    >>>    [2, 2, 2],
    >>>])
    >>> annotations = [
    >>>    ObjectAnnotation(value=Mask(mask=arr, color=1), name="dog"),
    >>>    ObjectAnnotation(value=Mask(mask=arr, color=2), name="cat"),
    >>>]

    Args:
         mask (MaskData): An object containing the actual mask, `MaskData` can
            be shared across multiple `Masks` to more efficiently store data
            for mutually exclusive segmentations.
         color (Tuple[uint8, uint8, uint8]): RGB color or a single value
            indicating the values of the class in the `MaskData`
    """

    mask: MaskData
    color: Union[Tuple[int, int, int], int]

    @property
    def geometry(self) -> Dict[str, Tuple[int, int, int]]:
        mask = self.draw(color=1)
        contours, hierarchy = cv2.findContours(
            image=mask, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_NONE
        )

        holes = []
        external_contours = []
        for i in range(len(contours)):
            if hierarchy[0, i, 3] != -1:
                # determined to be a hole based on contour hierarchy
                holes.append(contours[i])
            else:
                external_contours.append(contours[i])

        external_polygons = self._extract_polygons_from_contours(
            external_contours
        )
        holes = self._extract_polygons_from_contours(holes)

        if not external_polygons.is_valid:
            external_polygons = external_polygons.buffer(0)

        if not holes.is_valid:
            holes = holes.buffer(0)

        return external_polygons.difference(holes).__geo_interface__

    def draw(
        self,
        height: Optional[int] = None,
        width: Optional[int] = None,
        canvas: Optional[np.ndarray] = None,
        color: Optional[Union[int, Tuple[int, int, int]]] = None,
        thickness=None,
    ) -> np.ndarray:
        """Converts the Mask object into a numpy array

        Args:
            height (int): Optionally resize mask height before drawing.
            width (int): Optionally resize mask width before drawing.
            canvas (np.ndarray): Optionall provide a canvas to draw on
            color (Union[int, Tuple[int,int,int]]): Color to draw the canvas.
                Defaults to using the encoded color in the mask.
                int will return the mask as a 1d array
                tuple[int,int,int] will return the mask as a 3d array
            thickness (None): Unused, exists for a consistent interface.

        Returns:
            np.ndarray representing only this object
                as opposed to the mask that this object references which might have multiple objects determined by colors
        """
        mask = self.mask.value
        mask = np.all(mask == self.color, axis=2).astype(np.uint8)

        if height is not None or width is not None:
            mask = cv2.resize(
                mask, (width or mask.shape[1], height or mask.shape[0])
            )

        dims = [mask.shape[0], mask.shape[1]]
        color = color or self.color
        if isinstance(color, (tuple, list)):
            dims = dims + [len(color)]

        canvas = (
            canvas
            if canvas is not None
            else np.zeros(tuple(dims), dtype=np.uint8)
        )
        canvas[mask.astype(bool)] = color
        return canvas

    def _extract_polygons_from_contours(self, contours: List) -> MultiPolygon:
        contours = map(np.squeeze, contours)
        filtered_contours = filter(lambda contour: len(contour) > 2, contours)
        polygons = map(Polygon, filtered_contours)
        return MultiPolygon(polygons)

    def create_url(self, signer: Callable[[bytes], str]) -> str:
        """
        Update the segmentation mask to have a url.
        Only update the mask if it doesn't already have a url

        Args:
            signer: A function that accepts bytes and returns a signed url.
        Returns:
            the url for the mask
        """
        return self.mask.create_url(signer)

    @field_validator("color")
    def is_valid_color(cls, color):
        if isinstance(color, (tuple, list)):
            if len(color) == 1:
                color = [color[0]] * 3
            if len(color) != 3:
                raise ValueError(
                    "Segmentation colors must be either a (r,g,b) tuple or a single grayscale value"
                )
            elif not all([0 <= c <= 255 for c in color]):
                raise ValueError(
                    f"All rgb colors must be between 0 and 255. Found : {color}"
                )
        elif not (0 <= color <= 255):
            raise ValueError(
                f"All rgb colors must be between 0 and 255. Found : {color}"
            )

        return color



import abc
from uuid import UUID, uuid4
from typing import Any, Dict, Optional

from .feature import FeatureSchema
from pydantic import PrivateAttr, ConfigDict


class BaseAnnotation(FeatureSchema, abc.ABC):
    """Base annotation class. Shouldn't be directly instantiated"""

    _uuid: Optional[UUID] = PrivateAttr()
    extra: Dict[str, Any] = {}

    model_config = ConfigDict(extra="allow")

    def __init__(self, **data):
        super().__init__(**data)
        extra_uuid = data.get("extra", {}).get("uuid")
        self._uuid = data.get("_uuid") or extra_uuid or uuid4()



from datetime import datetime, timedelta
from labelbox.schema.search_filters import (
    IntegerValue,
    RangeDateTimeOperatorWithSingleValue,
    RangeOperatorWithSingleValue,
    DateRange,
    RangeOperatorWithValue,
    DateRangeValue,
    DateValue,
    IdOperator,
    OperationType,
    OrganizationFilter,
    TaskCompletedCountFilter,
    WorkforceRequestedDateFilter,
    WorkforceRequestedDateRangeFilter,
    WorkspaceFilter,
    TaskRemainingCountFilter,
)


def test_request_labeling_service_dashboard(requested_labeling_service):
    project, _ = requested_labeling_service

    try:
        project.get_labeling_service_dashboard()
    except Exception as e:
        assert False, f"An exception was raised: {e}"

    try:
        project.client.get_labeling_service_dashboards().get_one()
    except Exception as e:
        assert False, f"An exception was raised: {e}"


def test_request_labeling_service_dashboard_filters(requested_labeling_service):
    project, _ = requested_labeling_service

    organization = project.client.get_organization()
    org_filter = OrganizationFilter(
        operator=IdOperator.Is, values=[organization.uid]
    )

    try:
        project.client.get_labeling_service_dashboards(
            search_query=[org_filter]
        ).get_one()
    except Exception as e:
        assert False, f"An exception was raised: {e}"

    workforce_requested_filter_after = WorkforceRequestedDateFilter(
        operation=OperationType.WorforceRequestedDate,
        value=DateValue(
            operator=RangeDateTimeOperatorWithSingleValue.GreaterThanOrEqual,
            value=datetime.strptime("2024-01-01", "%Y-%m-%d"),
        ),
    )
    year_from_now = datetime.now() + timedelta(days=365)
    workforce_requested_filter_before = WorkforceRequestedDateFilter(
        operation=OperationType.WorforceRequestedDate,
        value=DateValue(
            operator=RangeDateTimeOperatorWithSingleValue.LessThanOrEqual,
            value=year_from_now,
        ),
    )

    try:
        project.client.get_labeling_service_dashboards(
            search_query=[
                workforce_requested_filter_after,
                workforce_requested_filter_before,
            ]
        ).get_one()
    except Exception as e:
        assert False, f"An exception was raised: {e}"

    workforce_date_range_filter = WorkforceRequestedDateRangeFilter(
        operation=OperationType.WorforceRequestedDate,
        value=DateRangeValue(
            operator=RangeOperatorWithValue.Between,
            value=DateRange(min="2024-01-01T00:00:00-0800", max=year_from_now),
        ),
    )

    try:
        project.client.get_labeling_service_dashboards(
            search_query=[workforce_date_range_filter]
        ).get_one()
    except Exception as e:
        assert False, f"An exception was raised: {e}"

    # with non existing data
    workspace_id = "clzzu4rme000008l42vnl4kre"
    workspace_filter = WorkspaceFilter(
        operation=OperationType.Workspace,
        operator=IdOperator.Is,
        values=[workspace_id],
    )
    labeling_service_dashboard = [
        ld
        for ld in project.client.get_labeling_service_dashboards(
            search_query=[workspace_filter]
        )
    ]
    assert len(labeling_service_dashboard) == 0
    assert labeling_service_dashboard == []

    task_done_count_filter = TaskCompletedCountFilter(
        operation=OperationType.TaskCompletedCount,
        value=IntegerValue(
            operator=RangeOperatorWithSingleValue.GreaterThanOrEqual, value=0
        ),
    )
    task_remaining_count_filter = TaskRemainingCountFilter(
        operation=OperationType.TaskRemainingCount,
        value=IntegerValue(
            operator=RangeOperatorWithSingleValue.GreaterThanOrEqual, value=0
        ),
    )

    try:
        project.client.get_labeling_service_dashboards(
            search_query=[task_done_count_filter, task_remaining_count_filter]
        ).get_one()
    except Exception as e:
        assert False, f"An exception was raised: {e}"



import pytest

from labelbox.data.annotation_types.classification.classification import (
    Checklist,
    ClassificationAnnotation,
    ClassificationAnswer,
    Radio,
)
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.rectangle import Rectangle
from labelbox.data.annotation_types.video import VideoObjectAnnotation


@pytest.fixture
def bbox_video_annotation_objects():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right
            ),
            classifications=[
                ClassificationAnnotation(
                    name="nested",
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="radio_option_1",
                            classifications=[
                                ClassificationAnnotation(
                                    name="nested_checkbox",
                                    value=Checklist(
                                        answer=[
                                            ClassificationAnswer(
                                                name="nested_checkbox_option_1"
                                            ),
                                            ClassificationAnswer(
                                                name="nested_checkbox_option_2"
                                            ),
                                        ]
                                    ),
                                )
                            ],
                        )
                    ),
                )
            ],
        ),
        VideoObjectAnnotation(
            name="bbox",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Rectangle(
                start=Point(x=186.0, y=98.0),  # Top left
                end=Point(x=490.0, y=341.0),  # Bottom right
            ),
        ),
    ]

    return bbox_annotation



import json
from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
import labelbox.types as lb_types
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import Label, ObjectAnnotation, Rectangle, Point

DATAROW_ID = "ckrb1sf1i1g7i0ybcdc6oc8ct"


def test_rectangle():
    with open("tests/data/assets/ndjson/rectangle_import.json", "r") as file:
        data = json.load(file)
    labels = [
        Label(
            data=GenericDataRowData(
                uid="ckrb1sf1i1g7i0ybcdc6oc8ct",
            ),
            annotations=[
                ObjectAnnotation(
                    name="bbox",
                    extra={
                        "uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72",
                    },
                    value=Rectangle(
                        start=Point(x=38.0, y=28.0),
                        end=Point(x=81.0, y=69.0),
                    ),
                )
            ],
        )
    ]
    res = list(NDJsonConverter.serialize(labels))
    assert res == data


def test_rectangle_inverted_start_end_points():
    with open("tests/data/assets/ndjson/rectangle_import.json", "r") as file:
        data = json.load(file)

    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=81, y=69),
            end=lb_types.Point(x=38, y=28),
        ),
        extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"},
    )

    label = lb_types.Label(data={"uid": DATAROW_ID}, annotations=[bbox])

    res = list(NDJsonConverter.serialize([label]))
    assert res == data

    expected_bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=38, y=28),
            end=lb_types.Point(x=81, y=69),
        ),
        extra={
            "uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72",
        },
    )

    label = lb_types.Label(
        data={"uid": DATAROW_ID}, annotations=[expected_bbox]
    )

    data = list(NDJsonConverter.serialize([label]))

    assert res == data


def test_rectangle_mixed_start_end_points():
    with open("tests/data/assets/ndjson/rectangle_import.json", "r") as file:
        data = json.load(file)

    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=81, y=28),
            end=lb_types.Point(x=38, y=69),
        ),
        extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"},
    )

    label = lb_types.Label(data={"uid": DATAROW_ID}, annotations=[bbox])

    res = list(NDJsonConverter.serialize([label]))
    assert res == data

    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=38, y=28),
            end=lb_types.Point(x=81, y=69),
        ),
        extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"},
    )

    label = lb_types.Label(data={"uid": DATAROW_ID}, annotations=[bbox])

    data = list(NDJsonConverter.serialize([label]))
    assert res == data


def test_benchmark_reference_label_flag_enabled():
    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=81, y=28),
            end=lb_types.Point(x=38, y=69),
        ),
        extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"},
    )

    label = lb_types.Label(
        data={"uid": DATAROW_ID},
        annotations=[bbox],
        is_benchmark_reference=True,
    )

    res = list(NDJsonConverter.serialize([label]))
    assert res[0]["isBenchmarkReferenceLabel"]


def test_benchmark_reference_label_flag_disabled():
    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=81, y=28),
            end=lb_types.Point(x=38, y=69),
        ),
        extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"},
    )

    label = lb_types.Label(
        data={"uid": DATAROW_ID},
        annotations=[bbox],
        is_benchmark_reference=False,
    )

    res = list(NDJsonConverter.serialize([label]))
    assert not res[0].get("isBenchmarkReferenceLabel")



from enum import Enum


class AnnotationImportState(Enum):
    """State of the import job when importing annotations (RUNNING, FAILED, or FINISHED).

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - RUNNING
         - Indicates that the import job is not done yet.
       * - FAILED
         - Indicates the import job failed. Check `AnnotationImport.errors` for more information
       * - FINISHED
         - Indicates the import job is no longer running. Check `AnnotationImport.statuses` for more information
    """

    RUNNING = "RUNNING"
    FAILED = "FAILED"
    FINISHED = "FINISHED"


class CollectionJobStatus(Enum):
    """Status of an asynchronous job over a collection.

    * - State
      - Description
    * - SUCCESS
      - Indicates job has successfully processed entire collection of data
    * - PARTIAL SUCCESS
      - Indicates some data in the collection has succeeded and other data have failed
    * - FAILURE
      - Indicates job has failed to process entire collection of data
    """

    SUCCESS = "SUCCESS"
    PARTIAL_SUCCESS = "PARTIAL SUCCESS"
    FAILURE = "FAILURE"



from typing import Optional, Callable, Dict, Any, List

from labelbox.adv_client import AdvClient
from pydantic import BaseModel, PrivateAttr


class EmbeddingVector(BaseModel):
    """
    A Vector Embedding for Custom Embedding.

    Attributes:
        embedding_id (str): The ID of the associated Embedding
        vector (list): The raw vector values - the number of entries should match the Embedding's dimensions
        clusters (list): The cluster groupings
    """

    embedding_id: str
    vector: List[float]
    clusters: Optional[List[int]] = None

    def to_gql(self) -> Dict[str, Any]:
        result = {"embeddingId": self.embedding_id, "vector": self.vector}
        if self.clusters:
            result["clusters"] = self.clusters
        return result


class Embedding(BaseModel):
    """
    An Embedding is used to power similarity search in Catalog.

    This model supports the representation of both `Precomputed` embeddings that Labelbox provides,
    and `Custom` embeddings which can be imported directly into Labelbox.

    Attributes:
        id (str): The ID of the embedding
        name (str): The name of the embedding
        dims (int): Refers to the size of the vector space in which words, phrases, or other entities are embedded
        custom (bool): Indicates whether the embedding is a Precomputed embedding or a Custom embedding
    """

    id: str
    name: str
    custom: bool
    dims: int
    _client: AdvClient = PrivateAttr()

    def __init__(self, client: AdvClient, **data):
        super().__init__(**data)
        self._client = client

    def delete(self):
        """
        Delete a custom embedding.  If the embedding does not exist or
        cannot be deleted, an AdvLibException is raised.
        """
        self._client.delete_embedding(self.id)

    def import_vectors_from_file(
        self,
        path: str,
        callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    ):
        """
        Import vectors into a given embedding from an NDJSON file.  An
        NDJSON file consists of newline delimited JSON.  Each line of the file
        is valid JSON, but the entire file itself is NOT.  The format of the
        file looks like:

            {"id": DATAROW ID1, "vector": [ array of floats ]}\n
            {"id": DATAROW ID2, "vector": [ array of floats ]}\n
            {"id": DATAROW ID3, "vector": [ array of floats ]}

        The vectors are added to the system in an async manner and it may take
        up to a couple minutes before they are usable via similarity search. Note
        that you also need to upload at least 1000 vectors in order for similarity
        search to be activated.

        Args:
            path: The path to the NDJSON file.
            callback: a callback function used get the status of each batch of lines uploaded.
        """
        self._client.import_vectors_from_file(self.id, path, callback)

    def get_imported_vector_count(self) -> int:
        """
        Return the # of vectors actually imported into Labelbox.  This will give you an accurate
        count of the number of vectors written into the vector search system.

        Returns:
            The number of imported vectors.
        """
        return self._client.get_imported_vector_count(self.id)

    @classmethod
    def type_name(cls):
        return cls.__name__.split(".")[-1]
